from __future__ import annotations
import json
from json import JSONEncoder, JSONDecoder
from typing import Type
from ._rfc7519.claims import (
    convert_claims,
    Claims,
    ClaimsOption,
    BaseClaimsRegistry,
    JWTClaimsRegistry,
)
from ._rfc7519.security import check_sensitive_data
from .jws import (
    JWSRegistry,
    serialize_compact,
    deserialize_compact,
)
from .jwe import (
    JWERegistry,
    encrypt_compact,
    decrypt_compact,
)
from .jwk import KeyFlexible
from .errors import InvalidPayloadError
from .util import to_bytes
from .registry import Header

__all__ = [
    # types
    "Claims",
    "ClaimsOption",
    # modules
    "BaseClaimsRegistry",
    "JWTClaimsRegistry",
    "Token",
    # methods
    "encode",
    "decode",
    "check_sensitive_data",
]


class Token:
    """The extracted token object, which contains ``header`` and ``claims``.

    :param header: the header part of the JWT
    :param claims: the payload part of the JWT
    """

    def __init__(self, header: Header, claims: Claims):
        #: header in dict
        self.header = header
        #: payload claims in dict
        self.claims = claims


def encode(
    header: Header,
    claims: Claims,
    key: KeyFlexible,
    algorithms: list[str] | None = None,
    registry: JWSRegistry | JWERegistry | None = None,
    encoder_cls: Type[JSONEncoder] | None = None,
    default_type: str | None = "JWT",
) -> str:
    """Encode a JSON Web Token with the given header, and claims.

    :param header: A dict of the JWT header
    :param claims: A dict of the JWT claims to be encoded
    :param key: key used to sign the signature
    :param algorithms: a list of allowed algorithms
    :param registry: a ``JWSRegistry`` or ``JWERegistry`` to use
    :param encoder_cls: A JSONEncoder subclass to use
    :param default_type: default value of the ``typ`` header parameter
    """
    if default_type is not None:
        _header = {"typ": default_type, **header}
    else:
        _header = {**header}
    payload = convert_claims(claims, encoder_cls)
    if isinstance(registry, JWERegistry):
        return encrypt_compact(_header, payload, key, algorithms, registry)
    else:
        return serialize_compact(_header, payload, key, algorithms, registry)


def decode(
    value: bytes | str,
    key: KeyFlexible,
    algorithms: list[str] | None = None,
    registry: JWSRegistry | JWERegistry | None = None,
    decoder_cls: Type[JSONDecoder] | None = None,
) -> Token:
    """Decode the JSON Web Token string with the given key, and validate
    it with the claims requests.

    :param value: text of the JWT
    :param key: key used to verify the signature
    :param algorithms: a list of allowed algorithms
    :param registry: a ``JWSRegistry`` or ``JWERegistry`` to use
    :param decoder_cls: A JSONDecoder subclass to use
    :raise BadSignatureError: when signature verification fails
    :raise InvalidPayloadError: when payload is not a valid JSON object
    """
    _value = to_bytes(value)
    header: Header
    payload: bytes
    if isinstance(registry, JWERegistry):
        header, payload = _decode_jwe(_value, key, algorithms, registry)
    else:
        header, payload = _decode_jws(_value, key, algorithms, registry)

    try:
        claims: Claims = json.loads(payload, cls=decoder_cls)
    except (TypeError, ValueError):
        raise InvalidPayloadError()

    return Token(header, claims)


def _decode_jwe(
    value: bytes, key: KeyFlexible, algorithms: list[str] | None = None, registry: JWERegistry | None = None
) -> tuple[Header, bytes]:
    jwe_obj = decrypt_compact(value, key, algorithms, registry)
    assert jwe_obj.plaintext is not None
    return jwe_obj.headers(), jwe_obj.plaintext


def _decode_jws(
    value: bytes, key: KeyFlexible, algorithms: list[str] | None = None, registry: JWSRegistry | None = None
) -> tuple[Header, bytes]:
    jws_obj = deserialize_compact(value, key, algorithms, registry)
    assert jws_obj.payload is not None
    return jws_obj.headers(), jws_obj.payload



from unittest import TestCase
from joserfc.jwe import (
    JWERegistry,
    GeneralJSONEncryption,
    GeneralJSONSerialization,
    encrypt_compact,
    decrypt_compact,
    encrypt_json,
    decrypt_json,
)
from joserfc.jwa import JWE_ENC_MODELS
from joserfc.jwk import KeySet
from joserfc.errors import InvalidEncryptionAlgorithmError
from joserfc.drafts.jwe_ecdh_1pu import JWE_ALG_MODELS, register_ecdh_1pu
from tests.base import TestFixture, load_key

register_ecdh_1pu()
ecdh_registry = JWERegistry(algorithms=[m.name for m in JWE_ALG_MODELS] + [enc.name for enc in JWE_ENC_MODELS])


class TestECDH1PUCompact(TestFixture):
    def run_test(self, data):
        alice_key = load_key("ec-p256-alice.json")
        bob_key = load_key("ec-p256-bob.json")
        value: str = data["value"]
        payload = data["payload"].encode("utf-8")
        obj = decrypt_compact(
            value.encode("utf-8"),
            private_key=bob_key,
            registry=ecdh_registry,
            sender_key=alice_key,
        )
        self.assertEqual(obj.protected["alg"], data["alg"])
        self.assertEqual(obj.protected["enc"], data["enc"])
        self.assertEqual(obj.plaintext, payload)

    def run_compact_case(self, alg: str, enc: str, recipient_key, sender_key):
        protected = {"alg": alg, "enc": enc}
        value = encrypt_compact(
            protected,
            b"hello",
            public_key=recipient_key,
            registry=ecdh_registry,
            sender_key=sender_key,
        )
        self.assertEqual(value.count("."), 4)
        obj = decrypt_compact(
            value,
            private_key=recipient_key,
            registry=ecdh_registry,
            sender_key=sender_key,
        )
        self.assertEqual(obj.plaintext, b"hello")

    def test_ecdh_1pu_compact_direct_mode(self):
        alice_key = load_key("ec-p256-alice.json")
        bob_key = load_key("ec-p256-bob.json")

        for enc in JWE_ENC_MODELS:
            self.run_compact_case("ECDH-1PU", enc.name, bob_key, alice_key)

        alice_key = load_key("okp-x25519-alice.json")
        bob_key = load_key("okp-x25519-bob.json")
        for enc in JWE_ENC_MODELS:
            self.run_compact_case("ECDH-1PU", enc.name, bob_key, alice_key)

    def test_ecdh_1pu_compact_agreement_mode(self):
        allowed_alg_values = [
            "ECDH-1PU+A128KW",
            "ECDH-1PU+A192KW",
            "ECDH-1PU+A256KW",
        ]
        allowed_enc_values = [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
        ]

        alice_key = load_key("ec-p256-alice.json")
        bob_key = load_key("ec-p256-bob.json")
        for alg in allowed_alg_values:
            for enc in allowed_enc_values:
                self.run_compact_case(alg, enc, bob_key, alice_key)

        alice_key = load_key("okp-x25519-alice.json")
        bob_key = load_key("okp-x25519-bob.json")
        for alg in allowed_alg_values:
            for enc in allowed_enc_values:
                self.run_compact_case(alg, enc, bob_key, alice_key)

    def test_ecdh_1pu_agreement_mode_with_other_encryption_algorithms(self):
        alice_key = load_key("ec-p256-alice.json")
        bob_key = load_key("ec-p256-bob.json")
        alg_values = [
            "ECDH-1PU+A128KW",
            "ECDH-1PU+A192KW",
            "ECDH-1PU+A256KW",
        ]
        other_enc_values = [
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]
        for alg in alg_values:
            for enc in other_enc_values:
                protected = {"alg": alg, "enc": enc}
                self.assertRaises(
                    InvalidEncryptionAlgorithmError,
                    encrypt_compact,
                    protected,
                    b"hello",
                    public_key=bob_key,
                    registry=ecdh_registry,
                    sender_key=alice_key,
                )

    def test_load_sender_key_via_skid(self):
        alice_key = load_key("ec-p256-alice.json", {"kid": "alice"})
        bob_key = load_key("ec-p256-bob.json", {"kid": "bob"})
        key = KeySet([alice_key, bob_key])

        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256", "kid": "bob", "skid": "alice"}
        value = encrypt_compact(
            protected,
            b"hello",
            public_key=key,
            registry=ecdh_registry,
            sender_key=key,
        )
        self.assertEqual(value.count("."), 4)
        obj = decrypt_compact(
            value,
            private_key=key,
            registry=ecdh_registry,
            sender_key=key,
        )
        self.assertEqual(obj.plaintext, b"hello")

    def test_sender_key_not_found_via_kid(self):
        alice_key = load_key("ec-p256-alice.json", {"kid": "alice"})
        bob_key = load_key("ec-p256-bob.json", {"kid": "bob"})
        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        value = encrypt_compact(
            protected,
            b"hello",
            public_key=bob_key,
            registry=ecdh_registry,
            sender_key=alice_key,
        )
        key_set = KeySet([alice_key, bob_key])
        self.assertRaises(
            ValueError,
            decrypt_compact,
            value,
            private_key=bob_key,
            registry=ecdh_registry,
            sender_key=key_set,
        )

    def test_sender_key_not_found_via_alg(self):
        alice_key = load_key("ec-p256-alice.json", {"kid": "alice"})
        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        bob_key = load_key("RFC7520-RSA-private.json")
        key_set = KeySet([bob_key])
        self.assertRaises(
            ValueError,
            encrypt_compact,
            protected,
            b"hello",
            public_key=alice_key,
            registry=ecdh_registry,
            sender_key=key_set,
        )


TestECDH1PUCompact.load_fixture("jwe_compact_ecdh_1pu.json")


class TestECDH1PUJSON(TestCase):
    def test_example_B(self):
        # https://datatracker.ietf.org/doc/html/draft-madden-jose-ecdh-1pu-04#appendix-B.11
        result: GeneralJSONSerialization = {
            "protected": (
                "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
                "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
                "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
                "RnFVQUZhMzlkeUJjIn19"
            ),
            "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"kid": "bob-key-2"},
                    "encrypted_key": (
                        "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CH"
                        "JQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                    ),
                },
                {
                    "header": {"kid": "2021-05-06"},
                    "encrypted_key": (
                        "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCq"
                        "RpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                    ),
                },
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
        }
        alice_key = load_key("okp-x25519-alice.json")
        bob_key = load_key("okp-x25519-bob.json", {"kid": "bob-key-2"})
        charlie_key = load_key("okp-x25519-charlie.json", {"kid": "2021-05-06"})
        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
            "epk": {"kty": "OKP", "crv": "X25519", "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc"},
        }
        obj = decrypt_json(
            result,
            KeySet([bob_key, charlie_key]),
            registry=ecdh_registry,
            sender_key=alice_key,
        )
        self.assertEqual(obj.protected, protected)
        value1 = encrypt_json(obj, KeySet([bob_key, charlie_key]), registry=ecdh_registry, sender_key=alice_key)
        self.assertEqual(value1["protected"], result["protected"])

    def test_with_key_set(self):
        alice_key = load_key("okp-x25519-alice.json", {"kid": "alice"})
        bob_key = load_key("okp-x25519-bob.json", {"kid": "bob"})
        charlie_key = load_key("okp-x25519-charlie.json", {"kid": "charlie"})
        keys = KeySet([alice_key, bob_key, charlie_key])

        obj = GeneralJSONEncryption(
            {"enc": "A128CBC-HS256"},
            plaintext=b"hello",
            aad=b"world",
        )
        obj.add_recipient({"alg": "ECDH-1PU+A128KW", "kid": "alice", "skid": "charlie"})
        obj.add_recipient({"alg": "ECDH-1PU+A256KW", "kid": "bob"})
        value = encrypt_json(obj, keys, registry=ecdh_registry, sender_key=keys)
        obj1 = decrypt_json(value, keys, registry=ecdh_registry, sender_key=keys)
        self.assertEqual(obj1.plaintext, b"hello")
        self.assertEqual(obj1.aad, b"world")



from unittest import TestCase
from joserfc.jwe import decrypt_compact, decrypt_json
from joserfc.jwk import RSAKey, OctKey, KeySet
from joserfc.util import json_b64encode, urlsafe_b64encode, to_bytes
from joserfc.jwe import JWERegistry, default_registry as registry
from joserfc.jwe import CompactEncryption, GeneralJSONEncryption
from joserfc._rfc7516.message import perform_encrypt
from joserfc._rfc7516.compact import represent_compact
from joserfc._rfc7516.json import represent_general_json
from joserfc.errors import UnsupportedAlgorithmError
from tests.base import load_key


class TestCompactExamples(TestCase):
    def test_A1(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.1
        # Example JWE using RSAES-OAEP and AES GCM
        plaintext = b"The true sign of intelligence is not knowledge but imagination."

        # A.1.1.  JOSE Header
        protected = {"alg": "RSA-OAEP", "enc": "A256GCM"}
        self.assertEqual(json_b64encode(protected), b"eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ")

        obj = CompactEncryption(protected, plaintext)

        # A.1.2.  Content Encryption Key (CEK)
        cek = bytes(
            [
                177,
                161,
                244,
                128,
                84,
                143,
                225,
                115,
                63,
                180,
                3,
                255,
                107,
                154,
                212,
                246,
                138,
                7,
                110,
                91,
                112,
                46,
                34,
                105,
                47,
                130,
                203,
                46,
                122,
                234,
                64,
                252,
            ]
        )

        # A.1.3.  Key Encryption
        key: RSAKey = load_key("RFC7516-A.1.3.json")
        obj.attach_recipient(key)

        enc = registry.get_enc(protected["enc"])

        # resulting encrypted key
        obj.recipient.encrypted_key = bytes(
            [
                56,
                163,
                154,
                192,
                58,
                53,
                222,
                4,
                105,
                218,
                136,
                218,
                29,
                94,
                203,
                22,
                150,
                92,
                129,
                94,
                211,
                232,
                53,
                89,
                41,
                60,
                138,
                56,
                196,
                216,
                82,
                98,
                168,
                76,
                37,
                73,
                70,
                7,
                36,
                8,
                191,
                100,
                136,
                196,
                244,
                220,
                145,
                158,
                138,
                155,
                4,
                117,
                141,
                230,
                199,
                247,
                173,
                45,
                182,
                214,
                74,
                177,
                107,
                211,
                153,
                11,
                205,
                196,
                171,
                226,
                162,
                128,
                171,
                182,
                13,
                237,
                239,
                99,
                193,
                4,
                91,
                219,
                121,
                223,
                107,
                167,
                61,
                119,
                228,
                173,
                156,
                137,
                134,
                200,
                80,
                219,
                74,
                253,
                56,
                185,
                91,
                177,
                34,
                158,
                89,
                154,
                205,
                96,
                55,
                18,
                138,
                43,
                96,
                218,
                215,
                128,
                124,
                75,
                138,
                243,
                85,
                25,
                109,
                117,
                140,
                26,
                155,
                249,
                67,
                167,
                149,
                231,
                100,
                6,
                41,
                65,
                214,
                251,
                232,
                87,
                72,
                40,
                182,
                149,
                154,
                168,
                31,
                193,
                126,
                215,
                89,
                28,
                111,
                219,
                125,
                182,
                139,
                235,
                195,
                197,
                23,
                234,
                55,
                58,
                63,
                180,
                68,
                202,
                206,
                149,
                75,
                205,
                248,
                176,
                67,
                39,
                178,
                60,
                98,
                193,
                32,
                238,
                122,
                96,
                158,
                222,
                57,
                183,
                111,
                210,
                55,
                188,
                215,
                206,
                180,
                166,
                150,
                166,
                106,
                250,
                55,
                229,
                72,
                40,
                69,
                214,
                216,
                104,
                23,
                40,
                135,
                212,
                28,
                127,
                41,
                80,
                175,
                174,
                168,
                115,
                171,
                197,
                89,
                116,
                92,
                103,
                246,
                83,
                216,
                182,
                176,
                84,
                37,
                147,
                35,
                45,
                219,
                172,
                99,
                226,
                233,
                73,
                37,
                124,
                42,
                72,
                49,
                242,
                35,
                127,
                184,
                134,
                117,
                114,
                135,
                206,
            ]
        )

        # A.1.4.  Initialization Vector
        iv = bytes([227, 197, 117, 252, 2, 219, 233, 68, 180, 225, 77, 219])
        obj.base64_segments["iv"] = urlsafe_b64encode(iv)

        # A.1.5.  Additional Authenticated Data
        aad = bytes(
            [
                101,
                121,
                74,
                104,
                98,
                71,
                99,
                105,
                79,
                105,
                74,
                83,
                85,
                48,
                69,
                116,
                84,
                48,
                70,
                70,
                85,
                67,
                73,
                115,
                73,
                109,
                86,
                117,
                89,
                121,
                73,
                54,
                73,
                107,
                69,
                121,
                78,
                84,
                90,
                72,
                81,
                48,
                48,
                105,
                102,
                81,
            ]
        )
        self.assertEqual(json_b64encode(protected), aad)
        obj.base64_segments["aad"] = aad

        # A.1.6.  Content Encryption
        ciphertext = bytes(
            [
                229,
                236,
                166,
                241,
                53,
                191,
                115,
                196,
                174,
                43,
                73,
                109,
                39,
                122,
                233,
                96,
                140,
                206,
                120,
                52,
                51,
                237,
                48,
                11,
                190,
                219,
                186,
                80,
                111,
                104,
                50,
                142,
                47,
                167,
                59,
                61,
                181,
                127,
                196,
                21,
                40,
                82,
                242,
                32,
                123,
                143,
                168,
                226,
                73,
                216,
                176,
                144,
                138,
                247,
                106,
                60,
                16,
                205,
                160,
                109,
                64,
                63,
                192,
            ]
        )
        r_ciphertext, r_tag = enc.encrypt(plaintext, cek, iv, aad)
        self.assertEqual(r_ciphertext, ciphertext)
        obj.base64_segments["ciphertext"] = urlsafe_b64encode(ciphertext)

        tag = bytes([92, 80, 104, 49, 133, 25, 161, 215, 173, 101, 219, 211, 136, 91, 210, 145])
        self.assertEqual(r_tag, tag)
        obj.base64_segments["tag"] = urlsafe_b64encode(r_tag)

        # A.1.7.  Complete Representation
        expected = (
            "eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ."
            "OKOawDo13gRp2ojaHV7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGe"
            "ipsEdY3mx_etLbbWSrFr05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDb"
            "Sv04uVuxIp5Zms1gNxKKK2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaV"
            "mqgfwX7XWRxv2322i-vDxRfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je8"
            "1860ppamavo35UgoRdbYaBcoh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi"
            "6UklfCpIMfIjf7iGdXKHzg."
            "48V1_ALb6US04U3b."
            "5eym8TW_c8SuK0ltJ3rpYIzOeDQz7TALvtu6UG9oMo4vpzs9tX_EFShS8iB7j6ji"
            "SdiwkIr3ajwQzaBtQD_A."
            "XFBoMYUZodetZdvTiFvSkQ"
        )
        self.assertEqual(represent_compact(obj), to_bytes(expected))

        jwe_data = decrypt_compact(expected, key)
        self.assertEqual(jwe_data.plaintext, plaintext)

    def test_A2(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.2
        plaintext = b"Live long and prosper."
        self.assertEqual(
            plaintext,
            bytes(
                [76, 105, 118, 101, 32, 108, 111, 110, 103, 32, 97, 110, 100, 32, 112, 114, 111, 115, 112, 101, 114, 46]
            ),
        )

        # A.2.1.  JOSE Header
        protected = {"alg": "RSA1_5", "enc": "A128CBC-HS256"}
        self.assertEqual(json_b64encode(protected), b"eyJhbGciOiJSU0ExXzUiLCJlbmMiOiJBMTI4Q0JDLUhTMjU2In0")
        obj = CompactEncryption(protected, plaintext)

        # A.2.2.  Content Encryption Key (CEK)
        cek = bytes(
            [
                4,
                211,
                31,
                197,
                84,
                157,
                252,
                254,
                11,
                100,
                157,
                250,
                63,
                170,
                106,
                206,
                107,
                124,
                212,
                45,
                111,
                107,
                9,
                219,
                200,
                177,
                0,
                240,
                143,
                156,
                44,
                207,
            ]
        )

        # A.2.3.  Key Encryption
        key: RSAKey = load_key("RFC7516-A.2.3.json")
        obj.attach_recipient(key)

        obj.recipient.encrypted_key = bytes(
            [
                80,
                104,
                72,
                58,
                11,
                130,
                236,
                139,
                132,
                189,
                255,
                205,
                61,
                86,
                151,
                176,
                99,
                40,
                44,
                233,
                176,
                189,
                205,
                70,
                202,
                169,
                72,
                40,
                226,
                181,
                156,
                223,
                120,
                156,
                115,
                232,
                150,
                209,
                145,
                133,
                104,
                112,
                237,
                156,
                116,
                250,
                65,
                102,
                212,
                210,
                103,
                240,
                177,
                61,
                93,
                40,
                71,
                231,
                223,
                226,
                240,
                157,
                15,
                31,
                150,
                89,
                200,
                215,
                198,
                203,
                108,
                70,
                117,
                66,
                212,
                238,
                193,
                205,
                23,
                161,
                169,
                218,
                243,
                203,
                128,
                214,
                127,
                253,
                215,
                139,
                43,
                17,
                135,
                103,
                179,
                220,
                28,
                2,
                212,
                206,
                131,
                158,
                128,
                66,
                62,
                240,
                78,
                186,
                141,
                125,
                132,
                227,
                60,
                137,
                43,
                31,
                152,
                199,
                54,
                72,
                34,
                212,
                115,
                11,
                152,
                101,
                70,
                42,
                219,
                233,
                142,
                66,
                151,
                250,
                126,
                146,
                141,
                216,
                190,
                73,
                50,
                177,
                146,
                5,
                52,
                247,
                28,
                197,
                21,
                59,
                170,
                247,
                181,
                89,
                131,
                241,
                169,
                182,
                246,
                99,
                15,
                36,
                102,
                166,
                182,
                172,
                197,
                136,
                230,
                120,
                60,
                58,
                219,
                243,
                149,
                94,
                222,
                150,
                154,
                194,
                110,
                227,
                225,
                112,
                39,
                89,
                233,
                112,
                207,
                211,
                241,
                124,
                174,
                69,
                221,
                179,
                107,
                196,
                225,
                127,
                167,
                112,
                226,
                12,
                242,
                16,
                24,
                28,
                120,
                182,
                244,
                213,
                244,
                153,
                194,
                162,
                69,
                160,
                244,
                248,
                63,
                165,
                141,
                4,
                207,
                249,
                193,
                79,
                131,
                0,
                169,
                233,
                127,
                167,
                101,
                151,
                125,
                56,
                112,
                111,
                248,
                29,
                232,
                90,
                29,
                147,
                110,
                169,
                146,
                114,
                165,
                204,
                71,
                136,
                41,
                252,
            ]
        )

        # A.2.4.  Initialization Vector
        iv = bytes([3, 22, 60, 12, 43, 67, 104, 105, 108, 108, 105, 99, 111, 116, 104, 101])
        obj.base64_segments["iv"] = urlsafe_b64encode(iv)
        self.assertEqual(obj.base64_segments["iv"], b"AxY8DCtDaGlsbGljb3RoZQ")

        # A.2.5.  Additional Authenticated Data
        aad = bytes(
            [
                101,
                121,
                74,
                104,
                98,
                71,
                99,
                105,
                79,
                105,
                74,
                83,
                85,
                48,
                69,
                120,
                88,
                122,
                85,
                105,
                76,
                67,
                74,
                108,
                98,
                109,
                77,
                105,
                79,
                105,
                74,
                66,
                77,
                84,
                73,
                52,
                81,
                48,
                74,
                68,
                76,
                85,
                104,
                84,
                77,
                106,
                85,
                50,
                73,
                110,
                48,
            ]
        )
        self.assertEqual(json_b64encode(protected), aad)
        obj.base64_segments["aad"] = aad

        # A.2.6.  Content Encryption
        ciphertext = bytes(
            [
                40,
                57,
                83,
                181,
                119,
                33,
                133,
                148,
                198,
                185,
                243,
                24,
                152,
                230,
                6,
                75,
                129,
                223,
                127,
                19,
                210,
                82,
                183,
                230,
                168,
                33,
                215,
                104,
                143,
                112,
                56,
                102,
            ]
        )
        enc = registry.get_enc(protected["enc"])
        r_ciphertext, r_tag = enc.encrypt(plaintext, cek, iv, aad)
        self.assertEqual(r_ciphertext, ciphertext)
        obj.base64_segments["ciphertext"] = urlsafe_b64encode(ciphertext)

        self.assertEqual(r_tag, bytes([246, 17, 244, 190, 4, 95, 98, 3, 231, 0, 115, 157, 242, 203, 100, 191]))
        obj.base64_segments["tag"] = urlsafe_b64encode(r_tag)

        expected = (
            "eyJhbGciOiJSU0ExXzUiLCJlbmMiOiJBMTI4Q0JDLUhTMjU2In0."
            "UGhIOguC7IuEvf_NPVaXsGMoLOmwvc1GyqlIKOK1nN94nHPoltGRhWhw7Zx0-kFm"
            "1NJn8LE9XShH59_i8J0PH5ZZyNfGy2xGdULU7sHNF6Gp2vPLgNZ__deLKxGHZ7Pc"
            "HALUzoOegEI-8E66jX2E4zyJKx-YxzZIItRzC5hlRirb6Y5Cl_p-ko3YvkkysZIF"
            "NPccxRU7qve1WYPxqbb2Yw8kZqa2rMWI5ng8OtvzlV7elprCbuPhcCdZ6XDP0_F8"
            "rkXds2vE4X-ncOIM8hAYHHi29NX0mcKiRaD0-D-ljQTP-cFPgwCp6X-nZZd9OHBv"
            "-B3oWh2TbqmScqXMR4gp_A."
            "AxY8DCtDaGlsbGljb3RoZQ."
            "KDlTtXchhZTGufMYmOYGS4HffxPSUrfmqCHXaI9wOGY."
            "9hH0vgRfYgPnAHOd8stkvw"
        )
        self.assertEqual(represent_compact(obj), to_bytes(expected))

        # RSA1_5 is not allowed by default
        self.assertRaises(UnsupportedAlgorithmError, decrypt_compact, expected, key)
        _registry = JWERegistry(algorithms=["RSA1_5", "A128CBC-HS256"])
        jwe_data = decrypt_compact(expected, key, registry=_registry)
        self.assertEqual(jwe_data.plaintext, plaintext)

    def test_A3(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.3
        plaintext = b"Live long and prosper."
        protected = {"alg": "A128KW", "enc": "A128CBC-HS256"}
        key = OctKey.import_key({"kty": "oct", "k": "GawgguFyGrWKav7AX4VKUg"})
        expected = (
            "eyJhbGciOiJBMTI4S1ciLCJlbmMiOiJBMTI4Q0JDLUhTMjU2In0."
            "6KB707dM9YTIgHtLvtgWQ8mKwboJW3of9locizkDTHzBC2IlrT1oOQ."
            "AxY8DCtDaGlsbGljb3RoZQ."
            "KDlTtXchhZTGufMYmOYGS4HffxPSUrfmqCHXaI9wOGY."
            "U0m_YmjN04DJvceFICbCVQ"
        )

        extract_data = decrypt_compact(expected, key)
        self.assertEqual(extract_data.protected, protected)
        self.assertEqual(extract_data.plaintext, plaintext)

    def test_A4(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.4
        # A.4.1.  JWE Per-Recipient Unprotected Headers
        recipient1 = {"alg": "RSA1_5", "kid": "2011-04-29"}
        recipient2 = {"alg": "A128KW", "kid": "7"}

        # The algorithm and key used for the first recipient are the same as
        # that used in Appendix A.2.
        key1: RSAKey = load_key("RFC7516-A.2.3.json", {"kid": "2011-04-29"})

        # The algorithm and key used for the second recipient are the same as
        # that used in Appendix A.3.
        key2 = OctKey.import_key({"kty": "oct", "k": "GawgguFyGrWKav7AX4VKUg"}, {"kid": "7"})
        keys = KeySet([key1, key2])

        # A.4.2.  JWE Protected Header
        protected = {"enc": "A128CBC-HS256"}
        self.assertEqual(json_b64encode(protected), b"eyJlbmMiOiJBMTI4Q0JDLUhTMjU2In0")

        # A.4.3.  JWE Shared Unprotected Header
        shared_header = {"jku": "https://server.example.com/keys.jwks"}

        # A.4.6.  Content Encryption
        plaintext = b"Live long and prosper."

        ciphertext = bytes(
            [
                40,
                57,
                83,
                181,
                119,
                33,
                133,
                148,
                198,
                185,
                243,
                24,
                152,
                230,
                6,
                75,
                129,
                223,
                127,
                19,
                210,
                82,
                183,
                230,
                168,
                33,
                215,
                104,
                143,
                112,
                56,
                102,
            ]
        )

        # A.4.7.  Complete JWE JSON Serialization Representation
        expected = {
            "protected": "eyJlbmMiOiJBMTI4Q0JDLUhTMjU2In0",
            "unprotected": {"jku": "https://server.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"alg": "RSA1_5", "kid": "2011-04-29"},
                    "encrypted_key": (
                        "UGhIOguC7IuEvf_NPVaXsGMoLOmwvc1GyqlIKOK1nN94nHPoltGRhWhw7Zx0-"
                        "kFm1NJn8LE9XShH59_i8J0PH5ZZyNfGy2xGdULU7sHNF6Gp2vPLgNZ__deLKx"
                        "GHZ7PcHALUzoOegEI-8E66jX2E4zyJKx-YxzZIItRzC5hlRirb6Y5Cl_p-ko3"
                        "YvkkysZIFNPccxRU7qve1WYPxqbb2Yw8kZqa2rMWI5ng8OtvzlV7elprCbuPh"
                        "cCdZ6XDP0_F8rkXds2vE4X-ncOIM8hAYHHi29NX0mcKiRaD0-D-ljQTP-cFPg"
                        "wCp6X-nZZd9OHBv-B3oWh2TbqmScqXMR4gp_A"
                    ),
                },
                {
                    "header": {"alg": "A128KW", "kid": "7"},
                    "encrypted_key": "6KB707dM9YTIgHtLvtgWQ8mKwboJW3of9locizkDTHzBC2IlrT1oOQ",
                },
            ],
            "iv": "AxY8DCtDaGlsbGljb3RoZQ",
            "ciphertext": "KDlTtXchhZTGufMYmOYGS4HffxPSUrfmqCHXaI9wOGY",
            "tag": "Mz-VPPyU4RlcuYv1IwIvzw",
        }

        _registry = JWERegistry(algorithms=["RSA1_5", "A128KW", "A128CBC-HS256"])
        jwe_data = decrypt_json(expected, keys, registry=_registry)
        self.assertEqual(jwe_data.plaintext, plaintext)
        self.assertEqual(jwe_data.protected, protected)
        self.assertEqual(jwe_data.unprotected, shared_header)
        self.assertEqual(jwe_data.bytes_segments["ciphertext"], ciphertext)
        self.assertEqual(jwe_data.recipients[0].header, recipient1)
        self.assertEqual(jwe_data.recipients[1].header, recipient2)

    def test_A4_perform(self):
        recipient1 = {"alg": "RSA1_5", "kid": "2011-04-29"}
        recipient2 = {"alg": "A128KW", "kid": "7"}
        protected = {"enc": "A128CBC-HS256"}
        shared_header = {"jku": "https://server.example.com/keys.jwks"}
        key1: RSAKey = load_key("RFC7516-A.2.3.json", {"kid": "2011-04-29"})
        key2 = OctKey.import_key({"kty": "oct", "k": "GawgguFyGrWKav7AX4VKUg"}, {"kid": "7"})
        payload = b"Live long and prosper."
        obj = GeneralJSONEncryption(protected, payload, shared_header)
        obj.add_recipient(recipient1, key1)
        obj.add_recipient(recipient2, key2)
        _registry = JWERegistry(algorithms=["RSA1_5", "A128KW", "A128CBC-HS256"])
        perform_encrypt(obj, _registry)
        expected = represent_general_json(obj)

        keys = KeySet([key1, key2])
        jwe_data = decrypt_json(expected, keys, registry=_registry)
        self.assertEqual(jwe_data.plaintext, payload)



from unittest import TestCase
from cryptography.hazmat.primitives import hashes
from joserfc.jwk import ECKey, OctKey
from joserfc.errors import (
    InvalidExchangeKeyError,
    InvalidKeyTypeError,
    InvalidKeyCurveError,
)
from tests.keys import read_key


class TestECKey(TestCase):
    default_key = ECKey.generate_key()

    def test_exchange_derive_key(self):
        key1 = ECKey.generate_key("P-256")
        key2 = ECKey.generate_key("P-384")
        self.assertRaises(InvalidExchangeKeyError, key1.exchange_derive_key, key2)

        key3 = ECKey.generate_key("P-256", private=False)
        self.assertRaises(InvalidExchangeKeyError, key3.exchange_derive_key, key1)

    def run_import_key(self, name):
        public_pem = read_key(f"ec-{name}-public.pem")
        key1 = ECKey.import_key(public_pem)
        self.assertFalse(key1.is_private)

        private_pem = read_key(f"ec-{name}-private.pem")
        key2 = ECKey.import_key(private_pem)
        self.assertTrue(key2.is_private)

        # public key match
        self.assertEqual(
            key2.as_bytes(private=False),
            key1.as_bytes(private=False),
        )

        self.assertNotIn("d", key1.dict_value)
        self.assertIn("d", key2.dict_value)

    def test_import_p256_key(self):
        self.run_import_key("p256")

    def test_import_p384_key(self):
        self.run_import_key("p384")

    def test_import_p512_key(self):
        self.run_import_key("p512")

    def test_import_secp256k1_key(self):
        self.run_import_key("secp256k1")

    def test_import_from_native_keys(self):
        key = ECKey.generate_key()
        self.assertEqual(key, ECKey.import_key(key.private_key))

    def test_generate_key(self):
        self.assertRaises(InvalidKeyCurveError, ECKey.generate_key, "Invalid")

        key = ECKey.generate_key(private=True)
        self.assertTrue(key.is_private)
        self.assertIsNone(key.kid)

        key = ECKey.generate_key(private=False)
        self.assertFalse(key.is_private)
        self.assertIsNone(key.private_key)

        key = ECKey.generate_key(auto_kid=True)
        self.assertIsNotNone(key.kid)

    def test_import_from_der_bytes(self):
        value1 = self.default_key.as_der()
        value2 = self.default_key.as_der(private=False)

        key1 = ECKey.import_key(value1)
        key2 = ECKey.import_key(value2)
        self.assertEqual(value1, key1.as_der())
        self.assertEqual(value2, key1.as_der(private=False))
        self.assertEqual(value2, key2.as_der())

    def test_import_invalid_pem_key(self):
        private_pem = read_key("rsa-openssl-private.pem")
        self.assertRaises(InvalidKeyTypeError, ECKey.import_key, private_pem)

    def test_output_with_password(self):
        key = ECKey.import_key(read_key("ec-p256-private.pem"))
        pem = key.as_pem(password="secret")
        self.assertRaises(TypeError, ECKey.import_key, pem)
        key2 = ECKey.import_key(pem, password="secret")
        self.assertEqual(key.as_dict(), key2.as_dict())

    def test_key_eq(self):
        key1 = self.default_key
        key2 = ECKey.import_key(key1.as_dict())
        self.assertEqual(key1, key2)
        key3 = ECKey.generate_key()
        self.assertNotEqual(key1, key3)

    def test_key_eq_with_different_types(self):
        self.assertNotEqual(self.default_key, OctKey.generate_key())

    def test_derive_key_errors(self):
        self.assertRaises(InvalidKeyCurveError, ECKey.derive_key, "secret", "invalid")
        self.assertRaises(ValueError, ECKey.derive_key, "secret", kdf_name="invalid")

    def test_derive_key_with_default_kwargs(self):
        key1 = ECKey.derive_key("ec-secret-key")
        key2 = ECKey.derive_key("ec-secret-key")
        self.assertEqual(key1, key2)

        for crv in ["P-256", "P-384", "P-521", "secp256k1"]:
            key1 = ECKey.derive_key("ec-secret-key", crv)
            key2 = ECKey.derive_key("ec-secret-key", crv)
            self.assertEqual(key1, key2)

        for crv in ["P-256", "P-384", "P-521", "secp256k1"]:
            key1 = ECKey.derive_key("ec-secret-key", crv, kdf_name="PBKDF2")
            key2 = ECKey.derive_key("ec-secret-key", crv, kdf_name="PBKDF2")
            self.assertEqual(key1, key2)

    def test_derive_key_with_new_salt(self):
        curves = ["P-256", "P-384", "P-521", "secp256k1"]
        for crv in curves:
            key1 = ECKey.derive_key("ec-secret-key", crv, kdf_options={"salt": b"salt"})
            key2 = ECKey.derive_key("ec-secret-key", crv, kdf_options={"salt": b"salt"})
            self.assertEqual(key1, key2)

    def test_derive_key_with_different_hash(self):
        key1 = ECKey.derive_key("ec-secret-key", "P-256", kdf_options={"algorithm": hashes.SHA256()})
        key2 = ECKey.derive_key("ec-secret-key", "P-256", kdf_options={"algorithm": hashes.SHA512()})
        self.assertNotEqual(key1, key2)



from unittest import TestCase
from joserfc import jwe
from joserfc.jwe import GeneralJSONEncryption, FlattenedJSONEncryption
from joserfc.jwk import KeySet, RSAKey, ECKey, OctKey
from joserfc.errors import (
    DecodeError,
    ConflictAlgorithmError,
    InvalidKeyTypeError,
)


class TestJWEJSON(TestCase):
    rsa_key = RSAKey.generate_key()
    ec_key = ECKey.generate_key()

    def test_multiple_recipients_with_key(self):
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "RSA-OAEP"}, self.rsa_key)
        obj.add_recipient({"alg": "ECDH-ES+A128KW"}, self.ec_key)
        value = jwe.encrypt_json(obj, None)
        self.assertIn("recipients", value)
        self.assertEqual(len(value["recipients"]), 2)

    def test_multiple_recipients_without_key(self):
        key1 = RSAKey.generate_key(parameters={"kid": "rsa"})
        key2 = ECKey.generate_key(parameters={"kid": "ec"})
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "RSA-OAEP", "kid": "rsa"})
        obj.add_recipient({"alg": "ECDH-ES+A128KW", "kid": "ec"})
        value = jwe.encrypt_json(obj, KeySet([key1, key2]))
        self.assertIn("recipients", value)
        self.assertEqual(len(value["recipients"]), 2)

    def test_multiple_recipients_with_direct_mode(self):
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "dir"}, OctKey.generate_key())
        obj.add_recipient({"alg": "RSA-OAEP"}, self.rsa_key)
        self.assertRaises(
            ConflictAlgorithmError,
            jwe.encrypt_json,
            obj,
            None,
        )

    def test_with_aad(self):
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i", aad=b"foo")
        obj.add_recipient({"alg": "RSA-OAEP"}, self.rsa_key)
        value = jwe.encrypt_json(obj, None)
        obj1 = jwe.decrypt_json(value, self.rsa_key)
        self.assertEqual(obj1.aad, b"foo")

    def test_decode_multiple_recipients(self):
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "RSA-OAEP"}, self.rsa_key)
        obj.add_recipient({"alg": "ECDH-ES+A128KW"}, self.ec_key)
        value = jwe.encrypt_json(obj, None)
        self.assertRaises(
            InvalidKeyTypeError,
            jwe.decrypt_json,
            value,
            self.rsa_key,
        )
        registry = jwe.JWERegistry(verify_all_recipients=False)
        obj1 = jwe.decrypt_json(value, self.rsa_key, registry=registry)
        self.assertEqual(obj1.plaintext, b"i")

        key3 = OctKey.generate_key()
        self.assertRaises(
            DecodeError,
            jwe.decrypt_json,
            value,
            key3,
            registry=registry,
        )

    def test_flattened_encryption(self):
        key = OctKey.generate_key(128)
        protected = {"enc": "A128CBC-HS256"}
        plaintext = b"hello world"
        obj0 = FlattenedJSONEncryption(protected, plaintext)
        obj0.add_recipient({"alg": "A128KW"})
        value = jwe.encrypt_json(obj0, key)
        obj1 = jwe.decrypt_json(value, key)
        self.assertEqual(obj1.plaintext, plaintext)

        obj2 = FlattenedJSONEncryption(protected, plaintext)
        obj2.add_recipient({"alg": "A128KW"}, key)
        value = jwe.encrypt_json(obj0, None)
        obj3 = jwe.decrypt_json(value, key)
        self.assertEqual(obj3.plaintext, plaintext)



import unittest

from joserfc.jws import JWSRegistry
from joserfc.jwk import OctKey, RSAKey, ECKey, OKPKey, KeySet


class JWSRegistryTest(unittest.TestCase):
    oct_key = OctKey.generate_key()
    rsa_key = RSAKey.generate_key()
    ec_key = ECKey.generate_key()
    okp_key = OKPKey.generate_key()

    def test_guess_recommended_algorithm(self):
        alg = JWSRegistry.guess_algorithm(self.oct_key, JWSRegistry.Strategy.RECOMMENDED)
        self.assertEqual(alg.name, "HS256")

        alg = JWSRegistry.guess_algorithm(self.rsa_key, JWSRegistry.Strategy.RECOMMENDED)
        self.assertEqual(alg.name, "RS256")

        alg = JWSRegistry.guess_algorithm(self.ec_key, JWSRegistry.Strategy.RECOMMENDED)
        self.assertEqual(alg.name, "ES256")

        alg = JWSRegistry.guess_algorithm(self.okp_key, JWSRegistry.Strategy.RECOMMENDED)
        self.assertEqual(alg, None)

    def test_guess_security_algorithm(self):
        alg = JWSRegistry.guess_algorithm(self.oct_key, JWSRegistry.Strategy.SECURITY)
        self.assertEqual(alg.name, "HS512")

        alg = JWSRegistry.guess_algorithm(self.rsa_key, JWSRegistry.Strategy.SECURITY)
        self.assertEqual(alg.name, "RS512")

        alg = JWSRegistry.guess_algorithm(self.ec_key, JWSRegistry.Strategy.SECURITY)
        self.assertEqual(alg.name, "ES256")

        ec521 = ECKey.generate_key("P-521")
        alg = JWSRegistry.guess_algorithm(ec521, JWSRegistry.Strategy.SECURITY)
        self.assertEqual(alg.name, "ES512")

        alg = JWSRegistry.guess_algorithm(self.okp_key, JWSRegistry.Strategy.SECURITY)
        self.assertEqual(alg.name, "EdDSA")

    def test_filter_algorithms_default_names(self):
        all_names = list(JWSRegistry.algorithms.keys())
        explicit = JWSRegistry.filter_algorithms(self.rsa_key, all_names)
        default = JWSRegistry.filter_algorithms(self.rsa_key)
        self.assertEqual(explicit, default)

    def test_filter_algorithms_ed25519(self):
        """Ed25519 keys should only be compatible with EdDSA and Ed25519, not Ed448."""
        ed25519_key = OKPKey.generate_key("Ed25519")
        algs = JWSRegistry.filter_algorithms(ed25519_key)
        names = [alg.name for alg in algs]
        self.assertIn("EdDSA", names)
        self.assertIn("Ed25519", names)
        self.assertNotIn("Ed448", names)

    def test_filter_algorithms_ed448(self):
        """Ed448 keys should only be compatible with EdDSA and Ed448, not Ed25519."""
        ed448_key = OKPKey.generate_key("Ed448")
        algs = JWSRegistry.filter_algorithms(ed448_key)
        names = [alg.name for alg in algs]
        self.assertIn("EdDSA", names)
        self.assertIn("Ed448", names)
        self.assertNotIn("Ed25519", names)

    def test_filter_algorithms_with_key_set(self):
        """filter_algorithms should support KeySet and combine algorithms from all keys."""
        rsa_key1 = RSAKey.generate_key()
        rsa_key2 = RSAKey.generate_key()
        ec_key = ECKey.generate_key("P-256")
        key_set = KeySet([rsa_key1, rsa_key2, ec_key])

        algs = JWSRegistry.filter_algorithms(key_set, JWSRegistry.algorithms.keys())
        names = [alg.name for alg in algs]

        self.assertIn("RS256", names)
        self.assertIn("ES256", names)
        self.assertNotIn("ES384", names)
        self.assertEqual(names.count("RS256"), 1)

    def test_guess_algorithm_with_key_set(self):
        """guess_algorithm should find the best algorithm across all keys in the KeySet."""
        rsa_key = RSAKey.generate_key()
        ec_key = ECKey.generate_key("P-256")
        key_set = KeySet([rsa_key, ec_key])

        # RS256 comes before ES256 in the recommended list
        alg = JWSRegistry.guess_algorithm(key_set, JWSRegistry.Strategy.RECOMMENDED)
        self.assertEqual(alg.name, "RS256")

        # RS512 has the highest algorithm_security (512) among available algorithms
        alg = JWSRegistry.guess_algorithm(key_set, JWSRegistry.Strategy.SECURITY)
        self.assertEqual(alg.name, "RS512")



from unittest import TestCase
from joserfc.jws import (
    HeaderDict,
    FlattenedJSONSerialization,
    GeneralJSONSerialization,
    JWSRegistry,
    serialize_json,
    deserialize_json,
    detach_content,
)
from joserfc.jwk import RSAKey, OctKey, KeySet
from joserfc.errors import (
    DecodeError,
    BadSignatureError,
    UnsupportedHeaderError,
    UnsupportedAlgorithmError,
    ExceededSizeError,
)
from joserfc.util import json_b64encode, urlsafe_b64encode, to_str
from tests.base import load_key


class TestJSON(TestCase):
    key = OctKey.import_key("secret")

    def test_serialize_json(self):
        key: RSAKey = load_key("rsa-openssl-private.pem")
        member: HeaderDict = {"protected": {"alg": "RS256"}}

        # flattened
        value = serialize_json(member, b"hello", key)
        self.assertIn("signature", value)
        self.assertNotIn("signatures", value)

        obj = deserialize_json(value, key)
        self.assertEqual(obj.payload, b"hello")
        self.assertEqual(obj.headers(), {"alg": "RS256"})

        # general
        value = serialize_json([member], b"hello", key)
        self.assertNotIn("signature", value)
        self.assertIn("signatures", value)

        obj = deserialize_json(value, key)
        self.assertEqual(obj.payload, b"hello")

    def test_serialize_with_unprotected_header(self):
        key: RSAKey = load_key("rsa-openssl-private.pem")
        member: HeaderDict = {"protected": {"alg": "RS256"}, "header": {"kid": "alice"}}
        value = serialize_json(member, b"hello", key)
        self.assertIn("header", value)
        self.assertEqual(value["header"], member["header"])

        value = serialize_json([member], b"hello", key)
        self.assertIn("signatures", value)

        value = value["signatures"][0]
        self.assertIn("header", value)
        self.assertEqual(value["header"], member["header"])

    def test_use_key_set(self):
        key1 = load_key("ec-p256-alice.json", {"kid": "alice"})
        key2 = load_key("ec-p256-bob.json", {"kid": "bob"})
        keys = KeySet([key1, key2])
        member: HeaderDict = {"protected": {"alg": "ES256"}}
        value = serialize_json(member, b"hello", keys)
        self.assertIn("header", value)
        self.assertIn("kid", value["header"])

        # this will always pick alice key
        member = {"protected": {"alg": "ES256"}, "header": {"kid": "alice"}}
        value = serialize_json(member, b"hello", keys)
        self.assertEqual(value["header"], {"kid": "alice"})

        # header can also be an empty value
        member = {"protected": {"alg": "ES256"}, "header": {}}
        value = serialize_json(member, b"hello", keys)
        self.assertIn("kid", value["header"])

    def test_detach_content(self):
        member: HeaderDict = {"protected": {"alg": "ES256"}}
        key = load_key("ec-p256-alice.json")
        value = serialize_json(member, b"hello", key)
        self.assertIn("payload", value)
        new_value = detach_content(value)
        self.assertNotIn("payload", new_value)
        # detach again will not raise error
        detach_content(new_value)

    def test_invalid_payload(self):
        member: HeaderDict = {"protected": {"alg": "ES256"}}
        key = load_key("ec-p256-alice.json")
        value = serialize_json(member, b"hello", key)
        value["payload"] = "a"
        self.assertRaises(DecodeError, deserialize_json, value, key)
        value = serialize_json([member], b"hello", key)
        value["payload"] = "a"
        self.assertRaises(DecodeError, deserialize_json, value, key)

    def test_bad_signature(self):
        member: HeaderDict = {"protected": {"alg": "ES256"}}
        key1 = load_key("ec-p256-alice.json")
        key2 = load_key("ec-p256-bob.json")
        value = serialize_json(member, b"hello", key1)
        self.assertRaises(BadSignatureError, deserialize_json, value, key2)
        value = serialize_json([member], b"hello", key1)
        self.assertRaises(BadSignatureError, deserialize_json, value, key2)

    def test_with_public_header(self):
        key: RSAKey = load_key("rsa-openssl-private.pem")
        member: HeaderDict = {"header": {"alg": "RS256", "kid": "abc"}}
        value = serialize_json(member, b"hello", key)
        self.assertIn("header", value)
        obj = deserialize_json(value, key)
        self.assertEqual(obj.payload, b"hello")
        self.assertEqual(obj.headers(), {"alg": "RS256", "kid": "abc"})

    def test_strict_check_header(self):
        member: HeaderDict = {"protected": {"alg": "HS256", "custom": "hi"}}
        self.assertRaises(UnsupportedHeaderError, serialize_json, member, b"hi", self.key)
        registry = JWSRegistry(strict_check_header=False)
        serialize_json(member, b"hi", self.key, registry=registry)

    def test_unsupported_algorithm(self):
        member: HeaderDict = {"protected": {"alg": "HS256"}}
        value = serialize_json(member, b"hi", self.key)
        registry = JWSRegistry(algorithms=["HS512"])
        self.assertRaises(UnsupportedAlgorithmError, deserialize_json, value, self.key, registry=registry)

    def test_prevent_overwrite_header(self):
        member: HeaderDict = {"protected": {"alg": "HS256", "kid": "a"}}
        value = serialize_json(member, b"hello", self.key)
        value["header"] = {"kid": "b"}
        obj = deserialize_json(value, self.key)
        self.assertEqual(obj.headers()["kid"], "a")

    def test_header_exceeded_size_error(self):
        data: FlattenedJSONSerialization = {
            "protected": to_str(json_b64encode({f"a{i}": "a" * i for i in range(100)})),
            "payload": to_str(json_b64encode({"a": "a"})),
            "signature": to_str(urlsafe_b64encode(b"o")),
        }
        self.assertRaises(ExceededSizeError, deserialize_json, data, self.key)

        data: GeneralJSONSerialization = {
            "payload": to_str(json_b64encode({"a": "a"})),
            "signatures": [
                {
                    "protected": to_str(json_b64encode({f"a{i}": "a" * i for i in range(100)})),
                    "signature": to_str(urlsafe_b64encode(b"o")),
                }
            ],
        }
        self.assertRaises(ExceededSizeError, deserialize_json, data, self.key)

    def test_payload_exceeded_size_error(self):
        data: FlattenedJSONSerialization = {
            "protected": to_str(json_b64encode({"alg": "HS256"})),
            "payload": to_str(json_b64encode({f"a{i}": "a" * i for i in range(1000)})),
            "signature": to_str(urlsafe_b64encode(b"o")),
        }
        self.assertRaises(ExceededSizeError, deserialize_json, data, self.key)

        data: GeneralJSONSerialization = {
            "payload": to_str(json_b64encode({f"a{i}": "a" * i for i in range(1000)})),
            "signatures": [
                {
                    "protected": to_str(json_b64encode({"alg": "HS256"})),
                    "signature": to_str(urlsafe_b64encode(b"o")),
                }
            ],
        }
        self.assertRaises(ExceededSizeError, deserialize_json, data, self.key)

    def test_signature_exceeded_size_error(self):
        data: FlattenedJSONSerialization = {
            "protected": to_str(json_b64encode({"alg": "HS256"})),
            "payload": to_str(json_b64encode({"a": "a"})),
            "signature": to_str(urlsafe_b64encode(("o" * 1000).encode("utf-8"))),
        }
        self.assertRaises(ExceededSizeError, deserialize_json, data, self.key)
        data: GeneralJSONSerialization = {
            "payload": to_str(json_b64encode({"a": "a"})),
            "signatures": [
                {
                    "protected": to_str(json_b64encode({"alg": "HS256"})),
                    "signature": to_str(urlsafe_b64encode(("o" * 1000).encode("utf-8"))),
                }
            ],
        }
        self.assertRaises(ExceededSizeError, deserialize_json, data, self.key)



from unittest import TestCase
from joserfc.jwk import OctKey
from joserfc.errors import SecurityWarning, KeyParameterError
from tests.keys import read_key


class TestOctKey(TestCase):
    def test_import_key_from_str(self):
        key = OctKey.import_key("rfc")
        self.assertEqual(key["k"], "cmZj")
        self.assertEqual(key.raw_value, b"rfc")
        self.assertEqual(dict(key), key.as_dict())

    def test_import_key_from_bytes(self):
        key = OctKey.import_key(b"rfc")
        self.assertEqual(key["k"], "cmZj")
        self.assertEqual(key.raw_value, b"rfc")
        self.assertEqual(dict(key), key.as_dict())

    def test_import_key_from_dict(self):
        # https://www.rfc-editor.org/rfc/rfc7517#appendix-A.3
        data = {
            "kty": "oct",
            "alg": "A128KW",
            "k": "GawgguFyGrWKav7AX4VKUg",
        }
        key = OctKey.import_key(data)
        self.assertEqual(key.as_dict(), data)

        # with use and key ops
        data = {
            "kty": "oct",
            "alg": "A128KW",
            "k": "GawgguFyGrWKav7AX4VKUg",
            "use": "sig",
            "key_ops": ["sign", "verify"],
        }
        key = OctKey.import_key(data)
        self.assertEqual(key.as_dict(), data)

    def test_thumbprint_uri(self):
        data = {
            "kty": "oct",
            "alg": "A128KW",
            "k": "GawgguFyGrWKav7AX4VKUg",
            "use": "sig",
            "key_ops": ["sign", "verify"],
        }
        key = OctKey.import_key(data)
        thumbprint = "k1JnWRfC-5zzmL72vXIuBgTLfVROXBakS4OmGcrMCoc"
        self.assertEqual(key.thumbprint_uri(), f"urn:ietf:params:oauth:jwk-thumbprint:sha-256:{thumbprint}")

    def test_import_missing_k(self):
        data = {
            "kty": "oct",
            "alg": "A128KW",
        }
        self.assertRaises(KeyParameterError, OctKey.import_key, data)

    def test_invalid_typeof_k(self):
        data = {
            "kty": "oct",
            "alg": "A128KW",
            "k": 123,
        }
        self.assertRaises(KeyParameterError, OctKey.import_key, data)

    def test_mismatch_use_key_ops(self):
        data = {"kty": "oct", "alg": "A128KW", "k": "GawgguFyGrWKav7AX4VKUg", "use": "sig", "key_ops": ["wrapKey"]}
        self.assertRaises(KeyParameterError, OctKey.import_key, data)

    def test_invalid_use(self):
        data = {
            "kty": "oct",
            "k": "GawgguFyGrWKav7AX4VKUg",
            "use": "invalid",
        }
        self.assertRaises(KeyParameterError, OctKey.import_key, data)

    def test_invalid_key_ops(self):
        data = {
            "kty": "oct",
            "k": "GawgguFyGrWKav7AX4VKUg",
            "key_ops": ["invalid"],
        }
        self.assertRaises(KeyParameterError, OctKey.import_key, data)

    def test_import_pem_key(self):
        public_pem = read_key("ec-p256-public.pem")
        self.assertWarns(SecurityWarning, OctKey.import_key, public_pem)

    def test_generate_key(self):
        key = OctKey.generate_key()
        self.assertEqual(len(key.raw_value), 32)
        self.assertIsNone(key.kid)

        key = OctKey.generate_key(None)
        self.assertEqual(len(key.raw_value), 32)
        self.assertIsNone(key.kid)

        self.assertRaises(ValueError, OctKey.generate_key, private=False)
        self.assertRaises(ValueError, OctKey.generate_key, 251)

        key = OctKey.generate_key(auto_kid=True)
        self.assertIsNotNone(key.kid)

    def test_generate_key_with_warnings(self):
        self.assertWarns(SecurityWarning, OctKey.generate_key, 16)

    def test_import_key_with_warnings(self):
        self.assertWarns(SecurityWarning, OctKey.import_key, b"rfc")

    def test_key_eq(self):
        key1 = OctKey.generate_key()
        key2 = OctKey.import_key(key1.as_dict())
        self.assertIsNot(key1, key2)
        self.assertEqual(key1, key2)
        key3 = OctKey.generate_key()
        self.assertNotEqual(key1, key3)



from unittest import TestCase

from joserfc import jws
from joserfc.jwk import (
    guess_key,
    import_key,
    generate_key,
    thumbprint_uri,
)
from joserfc.jwk import KeySet, OctKey, RSAKey, ECKey, OKPKey
from joserfc.errors import (
    UnsupportedKeyAlgorithmError,
    UnsupportedKeyUseError,
    UnsupportedKeyOperationError,
    InvalidKeyTypeError,
    MissingKeyTypeError,
    InvalidKeyIdError,
)
from tests.keys import read_key


class Guest:
    def __init__(self):
        self._headers = {"alg": "HS256"}

    def headers(self):
        return self._headers

    def set_kid(self, kid):
        self._headers["kid"] = kid


class TestKeyMethods(TestCase):
    def test_guess_callable_key(self):
        oct_key = OctKey.generate_key(parameters={"kid": "1"})
        rsa_key = RSAKey.generate_key(parameters={"kid": "2"})

        def rsa_key_func(obj):
            return rsa_key

        def key_set_func(obj):
            return KeySet([oct_key, rsa_key])

        key = guess_key(rsa_key_func, Guest())
        self.assertIsInstance(key, RSAKey)

        guest = Guest()
        guest.set_kid("2")
        key = guess_key(key_set_func, guest)
        self.assertIsInstance(key, RSAKey)

    def test_guess_key_set(self):
        key_set = KeySet([OctKey.generate_key(), RSAKey.generate_key()])
        guest = Guest()
        guest._headers["alg"] = "HS256"

        self.assertRaises(InvalidKeyIdError, guess_key, key_set, guest)
        key1 = guess_key(key_set, guest, True)
        self.assertIsInstance(key1, OctKey)
        guess_key(key_set, guest)

        guest = Guest()
        guest._headers["alg"] = "RS256"
        self.assertRaises(InvalidKeyIdError, guess_key, key_set, guest)
        key2 = guess_key(key_set, guest, True)
        self.assertIsInstance(key2, RSAKey)

        guest = Guest()
        guest._headers["alg"] = "ES256"
        self.assertRaises(ValueError, guess_key, key_set, guest, True)

    def test_invalid_key(self):
        self.assertRaises(ValueError, guess_key, {}, Guest())

    def test_import_key(self):
        # test bytes
        key = import_key(b"secret", "oct")
        self.assertIsInstance(key, OctKey)

        # test string
        key = import_key("secret", "oct")
        self.assertIsInstance(key, OctKey)

        # test dict
        data = key.as_dict()
        key = import_key(data)
        self.assertIsInstance(key, OctKey)

        self.assertRaises(InvalidKeyTypeError, import_key, "secret", "invalid")

    def test_generate_key(self):
        key = generate_key("oct")
        self.assertIsInstance(key, OctKey)

        key = generate_key("RSA")
        self.assertIsInstance(key, RSAKey)

        key = generate_key("EC")
        self.assertIsInstance(key, ECKey)

        key = generate_key("OKP")
        self.assertIsInstance(key, OKPKey)

        self.assertRaises(InvalidKeyTypeError, generate_key, "invalid", 8)

    def test_check_use(self):
        key = OctKey.import_key("secret", {"use": "sig"})
        key.check_use("sig")
        self.assertRaises(UnsupportedKeyUseError, key.check_use, "enc")
        self.assertRaises(UnsupportedKeyUseError, key.check_use, "invalid")

    def test_check_alg(self):
        key = OctKey.import_key("secret", {"alg": "HS256"})
        key.check_alg("HS256")
        self.assertRaises(UnsupportedKeyAlgorithmError, key.check_alg, "RS256")

    def test_alg_property(self):
        key = OctKey.import_key("secret")
        self.assertIsNone(key.alg)

        key = OctKey.import_key("secret", {"alg": "HS256"})
        self.assertEqual(key.alg, "HS256")

    def test_check_ops(self):
        key = OctKey.import_key("secret", {"key_ops": ["sign", "verify"]})
        key.check_key_op("sign")
        self.assertRaises(UnsupportedKeyOperationError, key.check_key_op, "wrapKey")
        self.assertRaises(UnsupportedKeyOperationError, key.check_key_op, "invalid")
        key = RSAKey.generate_key(private=False)
        self.assertRaises(UnsupportedKeyOperationError, key.check_key_op, "sign")

    def test_import_without_kty(self):
        self.assertRaises(MissingKeyTypeError, import_key, {})

    def test_thumbprint_uri(self):
        value = thumbprint_uri(
            {
                "kty": "EC",
                "crv": "P-256",
                "x": "jJ6Flys3zK9jUhnOHf6G49Dyp5hah6CNP84-gY-n9eo",
                "y": "nhI6iD5eFXgBTLt_1p3aip-5VbZeMhxeFSpjfEAf7Ww",
            }
        )
        expected = "urn:ietf:params:oauth:jwk-thumbprint:sha-256:w9eYdC6_s_tLQ8lH6PUpc0mddazaqtPgeC2IgWDiqY8"
        self.assertEqual(value, expected)

    def test_find_correct_key_with_use(self):
        key = OctKey.generate_key()
        dict_key = key.as_dict()

        key1 = OctKey.import_key(dict_key, {"use": "enc"})
        key2 = OctKey.import_key(dict_key, {"use": "sig"})
        self.assertEqual(key1.kid, key2.kid)

        key_set = KeySet([key1, key2])
        # pick randomly
        jws.serialize_compact({"alg": "HS256"}, "foo", key_set)
        # get by kid
        jws.serialize_compact({"alg": "HS256", "kid": key2.kid}, "foo", key_set)

        key_set = KeySet([key1, key2, key2])
        # return the first found key
        jws.serialize_compact({"alg": "HS256", "kid": key2.kid}, "foo", key_set)

    def test_find_correct_key_with_alg(self):
        key = OctKey.generate_key()
        dict_key = key.as_dict()

        key1 = OctKey.import_key(dict_key, {"alg": "HS256"})
        key2 = OctKey.import_key(dict_key, {"alg": "dir"})

        self.assertEqual(key1.kid, key2.kid)

        key_set = KeySet([key1, key2])
        # pick randomly
        jws.serialize_compact({"alg": "HS256"}, "foo", key_set)
        # get by kid
        jws.serialize_compact({"alg": "HS256", "kid": key2.kid}, "foo", key_set)

    def test_import_bytes_keys(self):
        # check by ssh types
        key = import_key(read_key("ssh-rsa-public.pem"))
        self.assertEqual(key.key_type, "RSA")
        key = import_key(read_key("ssh-ecdsa-public.pem"))
        self.assertEqual(key.key_type, "EC")
        key = import_key(read_key("ssh-ed25519-public.pem"))
        self.assertEqual(key.key_type, "OKP")
        key = import_key("ssh-rsa-oct")
        self.assertEqual(key.key_type, "oct")

        # check by pem types
        key = import_key(read_key("rsa-openssl-public.pem"))
        self.assertEqual(key.key_type, "RSA")
        key = import_key(read_key("ec-p256-public.pem"))
        self.assertEqual(key.key_type, "EC")
        key = import_key(read_key("okp-ed448-public.pem"))
        self.assertEqual(key.key_type, "OKP")

        key = import_key(b"oct-key")
        self.assertEqual(key.key_type, "oct")



from unittest import TestCase
from joserfc.jws import (
    JWSRegistry,
    serialize_compact,
    deserialize_compact,
    detach_content,
)
from joserfc.jwk import OctKey, RSAKey, KeySet
from joserfc.errors import (
    BadSignatureError,
    DecodeError,
    MissingAlgorithmError,
    UnsupportedAlgorithmError,
    UnsupportedHeaderError,
    ExceededSizeError,
)
from joserfc.util import urlsafe_b64encode, json_b64encode


class TestCompact(TestCase):
    key = OctKey.import_key("secret")

    def test_registry_is_none(self):
        value = serialize_compact({"alg": "HS256"}, b"foo", self.key)
        expected = "eyJhbGciOiJIUzI1NiJ9.Zm9v.0pehoi-RMZM1jl-4TP_C4Y6BJ-bcmsuzfDyQpkpJkh0"
        self.assertEqual(value, expected)

        obj = deserialize_compact(value, self.key)
        self.assertEqual(obj.payload, b"foo")

    def test_bad_signature_error(self):
        key = OctKey.import_key("incorrect")
        value = b"eyJhbGciOiJIUzI1NiJ9.Zm9v.0pehoi-RMZM1jl-4TP_C4Y6BJ-bcmsuzfDyQpkpJkh0"
        self.assertRaises(BadSignatureError, deserialize_compact, value, key)

    def test_raise_unsupported_algorithm_error(self):
        self.assertRaises(UnsupportedAlgorithmError, serialize_compact, {"alg": "HS512"}, b"foo", self.key)
        self.assertRaises(UnsupportedAlgorithmError, serialize_compact, {"alg": "NOT"}, b"foo", self.key)

    def test_invalid_length(self):
        self.assertRaises(DecodeError, deserialize_compact, b"a.b.c.d", self.key)

    def test_no_invalid_header(self):
        # invalid base64
        value = b"abc.Zm9v.0pehoi"
        self.assertRaises(DecodeError, deserialize_compact, value, self.key)

        # no alg value
        value = b"eyJhIjoiYiJ9.Zm9v.0pehoi"
        self.assertRaises(MissingAlgorithmError, deserialize_compact, value, self.key)

    def test_invalid_payload(self):
        value = b"eyJhbGciOiJIUzI1NiJ9.a$b.0pehoi"
        self.assertRaises(DecodeError, deserialize_compact, value, self.key)

    def test_header_exceeded_size_error(self):
        exceeded_header = json_b64encode({f"a{i}": f"a{i}" for i in range(1000)})
        other = urlsafe_b64encode(b"o")
        fake_jws = exceeded_header + b"." + other + b"." + other
        self.assertRaises(ExceededSizeError, deserialize_compact, fake_jws, self.key)

    def test_payload_exceeded_size_error(self):
        header = json_b64encode({"alg": "HS256"})
        exceeded_payload = urlsafe_b64encode(("o" * 10000).encode("utf8"))
        fake_jws = header + b"." + exceeded_payload + b"." + urlsafe_b64encode(b"o")
        self.assertRaises(ExceededSizeError, deserialize_compact, fake_jws, self.key)

    def test_signature_exceeded_size_error(self):
        header = json_b64encode({"alg": "HS256"})
        exceeded_signature = urlsafe_b64encode(("o" * 1000).encode("utf8"))
        fake_jws = header + b"." + urlsafe_b64encode(b"o") + b"." + exceeded_signature
        self.assertRaises(ExceededSizeError, deserialize_compact, fake_jws, self.key)

    def test_with_key_set(self):
        keys = KeySet(
            [
                OctKey.import_key("a"),
                OctKey.import_key("b"),
                OctKey.import_key("c"),
            ]
        )
        value = serialize_compact({"alg": "HS256"}, b"foo", keys)
        obj = deserialize_compact(value, keys)
        self.assertEqual(obj.payload, b"foo")

        keys.keys.append(RSAKey.generate_key(auto_kid=True))
        value = serialize_compact({"alg": "RS256"}, b"foo", keys)
        obj = deserialize_compact(value, keys)
        self.assertEqual(obj.payload, b"foo")

    def test_strict_check_header(self):
        header = {"alg": "HS256", "custom": "hi"}
        self.assertRaises(UnsupportedHeaderError, serialize_compact, header, b"hi", self.key)

        registry = JWSRegistry(strict_check_header=False)
        serialize_compact(header, b"hi", self.key, registry=registry)

    def test_non_canonical_signature_encoding(self):
        text = "eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyIjoiYWRtaW4ifQ.VI29GgHzuh2xfF0bkRYvZIsSuQnbTXSIvuRyt7RDrwo"[:-1] + "p"
        self.assertRaises(BadSignatureError, deserialize_compact, text, OctKey.import_key("secret"))

    def test_detached_content(self):
        value = detach_content(serialize_compact({"alg": "HS256"}, b"foo", self.key))
        expected = "eyJhbGciOiJIUzI1NiJ9..0pehoi-RMZM1jl-4TP_C4Y6BJ-bcmsuzfDyQpkpJkh0"
        self.assertEqual(value, expected)
        obj = deserialize_compact(value, self.key, payload=b"foo")
        self.assertEqual(obj.payload, b"foo")



from joserfc.jwk import OctKey
from joserfc.errors import (
    DecodeError,
    MissingAlgorithmError,
    BadSignatureError,
    InvalidHeaderValueError,
    MissingCritHeaderError,
    UnsupportedHeaderError,
)
from joserfc.util import to_bytes
from joserfc.jws import HeaderDict
from joserfc import jws
from tests.base import TestFixture


default_key = OctKey.import_key(
    {"kty": "oct", "k": ("AyM1SysPpbyDfgZld3umj1qzKObwVMkoqQ-EstJQLr_T-1qS0gZH75aKtMN3Yj0iPS4hcgUuTwjAzZr1Z9CAow")}
)


class TestRFC7797(TestFixture):
    def run_test(self, data):
        protected = data["protected"]
        payload = data["payload"]
        value1 = jws.serialize_compact(protected, payload, default_key)
        self.assertEqual(value1, data["compact"])
        obj1 = jws.deserialize_compact(value1, default_key, payload=payload)
        self.assertEqual(obj1.headers(), protected)
        self.assertEqual(obj1.payload, to_bytes(payload))
        value2 = jws.serialize_json({"protected": protected}, payload, default_key)
        self.assertEqual(value2, data["flattened_json"])
        obj2 = jws.deserialize_json(value2, default_key)
        self.assertTrue(obj2.flattened)
        self.assertEqual(obj2.payload, to_bytes(payload))
        self.assertEqual(obj2.members[0].protected, protected)

    def test_b64_without_crit(self):
        protected = {"alg": "HS256", "b64": False}
        self.assertRaises(MissingCritHeaderError, jws.serialize_compact, protected, "i", default_key)

    def test_invalid_b64_value(self):
        protected = {"alg": "HS256", "b64": "true", "crit": ["b64"]}
        self.assertRaises(InvalidHeaderValueError, jws.serialize_compact, protected, "i", default_key)

    def test_compact_invalid_value_length(self):
        self.assertRaises(DecodeError, jws.deserialize_compact, b"a.b.c.d.e", default_key)

    def test_invalid_header(self):
        self.assertRaises(DecodeError, jws.deserialize_compact, b"a.b.c", default_key)

    def test_compact_missing_alg(self):
        self.assertRaises(MissingAlgorithmError, jws.deserialize_compact, b"e30.a.b", default_key)

    def test_compact_bad_signature(self):
        protected = {"alg": "HS256", "b64": False, "crit": ["b64"]}
        value = jws.serialize_compact(protected, "hello", default_key)
        key2 = OctKey.import_key("secret")
        self.assertRaises(BadSignatureError, jws.deserialize_compact, value, key2)

    def test_compact_use_registry(self):
        protected = {"alg": "HS256", "b64": True, "crit": ["b64"]}
        value = jws.serialize_compact(protected, "hello", default_key)
        obj = jws.deserialize_compact(value, default_key)
        self.assertEqual(obj.protected, protected)

        protected = {"alg": "HS256"}
        value = jws.serialize_compact(protected, "hello", default_key)
        obj = jws.deserialize_compact(value, default_key)
        self.assertEqual(obj.protected, protected)

    def test_json_without_protected_header(self):
        header = {"alg": "HS256", "b64": False, "crit": ["b64"]}
        member: HeaderDict = {"header": header}
        self.assertRaises(UnsupportedHeaderError, jws.serialize_json, member, "hello", default_key)

    def test_general_json(self):
        member: HeaderDict = {"protected": {"alg": "HS256"}}
        value = jws.serialize_json([member], "hello", default_key)
        obj = jws.deserialize_json(value, default_key)
        self.assertFalse(obj.flattened)

    def test_json_bad_signature(self):
        member: HeaderDict = {"protected": {"alg": "HS256", "b64": False, "crit": ["b64"]}}
        value = jws.serialize_json(member, "hello", default_key)
        key2 = OctKey.import_key("secret")
        self.assertRaises(BadSignatureError, jws.deserialize_json, value, key2)


TestRFC7797.load_fixture("jws_rfc7797.json")



from __future__ import annotations
import typing as t
import secrets
from abc import ABCMeta, abstractmethod
from ..registry import Header, HeaderRegistryDict
from ..errors import InvalidKeyTypeError, InvalidKeyLengthError
from .._keys import Key, ECKey, OctKey

__all__ = [
    "Recipient",
    "CompactEncryption",
    "BaseJSONEncryption",
    "GeneralJSONEncryption",
    "FlattenedJSONEncryption",
    "JWEEncModel",
    "JWEZipModel",
    "KeyManagement",
    "JWEDirectEncryption",
    "JWEKeyEncryption",
    "JWEKeyWrapping",
    "JWEKeyAgreement",
    "JWEAlgModel",
]

KeyType = t.TypeVar("KeyType")


class Recipient(t.Generic[KeyType]):
    def __init__(
        self,
        parent: t.Union["CompactEncryption", "GeneralJSONEncryption", "FlattenedJSONEncryption"],
        header: Header | None = None,
        recipient_key: KeyType | None = None,
    ):
        self.__parent = parent
        self.header = header
        self.recipient_key = recipient_key
        self.sender_key: t.Optional[KeyType] = None
        self.encrypted_key: t.Optional[bytes] = None
        self.ephemeral_key: t.Optional[KeyType] = None

    def headers(self) -> Header:
        rv: Header = {}
        if isinstance(self.__parent, BaseJSONEncryption) and self.__parent.unprotected:
            rv.update(self.__parent.unprotected)
        if self.header:
            rv.update(self.header)

        rv.update(self.__parent.protected)
        return rv

    def add_header(self, k: str, v: t.Any) -> None:
        if isinstance(self.__parent, CompactEncryption):
            self.__parent.protected.update({k: v})
        elif self.header:
            self.header.update({k: v})
        else:
            self.header = {k: v}

    def set_kid(self, kid: str) -> None:
        self.add_header("kid", kid)


class CompactEncryption:
    """An object to represent the JWE Compact Serialization. It is usually returned by
    ``decrypt_compact`` method.
    """

    def __init__(self, protected: Header, plaintext: bytes | None = None):
        #: protected header in dict
        self.protected = protected
        #: the plaintext in bytes
        self.plaintext = plaintext
        self.recipient: Recipient[t.Any] | None = None
        self.bytes_segments: dict[str, bytes] = {}  # store the decoded segments
        self.base64_segments: dict[str, bytes] = {}  # store the encoded segments

    def headers(self) -> Header:
        """Returns the protected header values in dict."""
        return self.protected

    def attach_recipient(self, key: Key, header: Header | None = None) -> None:
        """Add a recipient to the JWE Compact Serialization. Please add a key that
        comply with the given "alg" value.

        :param key: an instance of a key, e.g. (OctKey, RSAKey, ECKey, and etc)
        :param header: extra header in dict
        """
        recipient = Recipient(self, None, key)
        if header:
            self.protected.update(header)
        self.recipient = recipient

    @property
    def recipients(self) -> list[Recipient[t.Any]]:
        if self.recipient is not None:
            return [self.recipient]
        return []


class BaseJSONEncryption(metaclass=ABCMeta):
    #: represents if the object is in flatten syntax
    flattened: t.ClassVar[bool]
    #: protected header in dict
    protected: Header
    #: the plaintext in bytes
    plaintext: t.Optional[bytes]
    #: unprotected header in dict
    unprotected: t.Optional[Header]
    #: an optional additional authenticated data
    aad: t.Optional[bytes]
    #: a list of recipients
    recipients: list[Recipient[t.Any]]

    def __init__(
        self,
        protected: Header,
        plaintext: bytes | None = None,
        unprotected: Header | None = None,
        aad: bytes | None = None,
    ):
        self.protected = protected
        self.plaintext = plaintext
        self.unprotected = unprotected
        self.aad = aad
        self.recipients = []
        self.bytes_segments: dict[str, bytes] = {}  # store the decoded segments
        self.base64_segments: dict[str, bytes] = {}  # store the encoded segments

    @abstractmethod
    def add_recipient(self, header: Header | None = None, key: Key | None = None) -> None:
        """Add a recipient to the JWE JSON Serialization. Please add a key that
        comply with the "alg" to this recipient.

        :param header: recipient's own (unprotected) header
        :param key: an instance of a key, e.g. (OctKey, RSAKey, ECKey, and etc)
        """


class GeneralJSONEncryption(BaseJSONEncryption):
    """An object to represent the JWE General JSON Serialization. It is used by
    ``encrypt_json``, and it is usually returned by ``decrypt_json`` method.

    To construct an object of ``GeneralJSONEncryption``:

    .. code-block:: python

        protected = {"enc": "A128CBC-HS256"}
        plaintext = b"hello world"
        obj = GeneralJSONEncryption(protected, plaintext)
        # then add each recipient
        obj.add_recipient({"alg": "A128KW"})
    """

    flattened = False

    def add_recipient(self, header: Header | None = None, key: Key | None = None) -> None:
        recipient = Recipient(self, header, key)
        if key and key.kid:
            recipient.set_kid(key.kid)
        self.recipients.append(recipient)


class FlattenedJSONEncryption(BaseJSONEncryption):
    """An object to represent the JWE Flattened JSON Serialization. It is used by
    ``encrypt_json``, and it is usually returned by ``decrypt_json`` method.

    To construct an object of ``FlattenedJSONEncryption``:

    .. code-block:: python

        protected = {"enc": "A128CBC-HS256"}
        plaintext = b"hello world"
        obj = FlattenedJSONEncryption(protected, plaintext)
        # then add each recipient
        obj.add_recipient({"alg": "A128KW"})
    """

    flattened = True

    def add_recipient(self, header: Header | None = None, key: Key | None = None) -> None:
        recipient = Recipient(self, header, key)
        if key and key.kid:
            recipient.set_kid(key.kid)
        self.recipients = [recipient]


class JWEEncModel(metaclass=ABCMeta):
    name: str
    description: str
    recommended: bool = False
    algorithm_type: t.Literal["JWE"] = "JWE"
    algorithm_location: t.Literal["enc"] = "enc"

    iv_size: int
    cek_size: int

    def generate_cek(self) -> bytes:
        return secrets.token_bytes(self.cek_size // 8)

    def generate_iv(self) -> bytes:
        return secrets.token_bytes(self.iv_size // 8)

    def check_iv(self, iv: bytes) -> bytes:
        if len(iv) * 8 != self.iv_size:  # pragma: no cover
            raise ValueError("Invalid 'iv' size")
        return iv

    @abstractmethod
    def encrypt(self, plaintext: bytes, cek: bytes, iv: bytes, aad: bytes) -> tuple[bytes, bytes]:
        pass

    @abstractmethod
    def decrypt(self, ciphertext: bytes, tag: bytes, cek: bytes, iv: bytes, aad: bytes) -> bytes:
        pass


class JWEZipModel(metaclass=ABCMeta):
    name: str
    description: str
    recommended: bool = True
    algorithm_type: t.Literal["JWE"] = "JWE"
    algorithm_location: t.Literal["zip"] = "zip"

    @abstractmethod
    def compress(self, s: bytes) -> bytes:
        pass

    @abstractmethod
    def decompress(self, s: bytes) -> bytes:
        pass


class KeyManagement:
    name: str
    description: str
    recommended: bool = False
    key_size: int | None = None
    key_types: list[str]
    security_warning: str | None = None

    algorithm_type: t.Literal["JWE"] = "JWE"
    algorithm_location: t.Literal["alg"] = "alg"
    more_header_registry: HeaderRegistryDict = {}

    @property
    def direct_mode(self) -> bool:
        return self.key_size is None

    def check_key_type(self, key: Key) -> None:
        if key.key_type not in self.key_types:
            raise InvalidKeyTypeError()

    def prepare_recipient_header(self, recipient: Recipient[t.Any]) -> None:
        raise NotImplementedError()


class JWEDirectEncryption(KeyManagement, metaclass=ABCMeta):
    key_types = ["oct"]

    @abstractmethod
    def compute_cek(self, size: int, recipient: Recipient[OctKey]) -> bytes:
        pass


class JWEKeyEncryption(KeyManagement, metaclass=ABCMeta):
    @property
    def direct_mode(self) -> bool:
        return False

    @abstractmethod
    def encrypt_cek(self, cek: bytes, recipient: Recipient[t.Any]) -> bytes:
        pass

    @abstractmethod
    def decrypt_cek(self, recipient: Recipient[t.Any]) -> bytes:
        pass


class JWEKeyWrapping(KeyManagement, metaclass=ABCMeta):
    key_size: int
    key_types = ["oct"]

    @property
    def direct_mode(self) -> bool:
        return False

    def check_op_key(self, op_key: bytes) -> None:
        if len(op_key) * 8 != self.key_size:
            raise InvalidKeyLengthError(f"A key of size {self.key_size} bits MUST be used")

    @abstractmethod
    def wrap_cek(self, cek: bytes, key: bytes) -> bytes:
        pass

    @abstractmethod
    def unwrap_cek(self, ek: bytes, key: bytes) -> bytes:
        pass

    @abstractmethod
    def encrypt_cek(self, cek: bytes, recipient: Recipient[OctKey]) -> bytes:
        pass

    @abstractmethod
    def decrypt_cek(self, recipient: Recipient[OctKey]) -> bytes:
        pass


class JWEKeyAgreement(KeyManagement, metaclass=ABCMeta):
    key_types = ["EC", "OKP"]
    tag_aware: bool = False
    key_wrapping: t.Optional[JWEKeyWrapping]

    def prepare_ephemeral_key(self, recipient: Recipient[ECKey]) -> None:
        recipient_key = recipient.recipient_key
        assert recipient_key is not None
        self.check_key_type(recipient_key)
        if recipient.ephemeral_key is None:
            ephemeral_key = recipient_key.generate_key(recipient_key.curve_name, private=True)
            recipient.ephemeral_key = ephemeral_key
        recipient.add_header("epk", recipient.ephemeral_key.as_dict(private=False))

    @abstractmethod
    def encrypt_agreed_upon_key(self, enc: JWEEncModel, recipient: Recipient[ECKey]) -> bytes:
        pass

    @abstractmethod
    def decrypt_agreed_upon_key(self, enc: JWEEncModel, recipient: Recipient[ECKey]) -> bytes:
        pass

    def wrap_cek_with_auk(self, cek: bytes, key: bytes) -> bytes:
        assert self.key_wrapping is not None
        return self.key_wrapping.wrap_cek(cek, key)

    def unwrap_cek_with_auk(self, ek: bytes, key: bytes) -> bytes:
        assert self.key_wrapping is not None
        return self.key_wrapping.unwrap_cek(ek, key)

    def encrypt_agreed_upon_key_with_tag(self, enc: JWEEncModel, recipient: Recipient[ECKey], tag: bytes) -> bytes:
        raise NotImplementedError()

    def decrypt_agreed_upon_key_with_tag(self, enc: JWEEncModel, recipient: Recipient[ECKey], tag: bytes) -> bytes:
        raise NotImplementedError()


JWEAlgModel = t.Union[JWEKeyEncryption, JWEKeyWrapping, JWEKeyAgreement, JWEDirectEncryption]



from __future__ import annotations
import struct
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash
from ..registry import Header
from ..util import to_bytes, urlsafe_b64decode


__all__ = [
    "derive_key_for_concat_kdf",
    "u32be_len_input",
]


def derive_key_for_concat_kdf(
    shared_key: bytes, header: Header, cek_size: int, key_size: int | None, tag: bytes | None = None
) -> bytes:
    # PartyUInfo
    apu_info = u32be_len_input(header.get("apu"), True)
    # PartyVInfo
    apv_info = u32be_len_input(header.get("apv"), True)
    # SuppPubInfo

    if key_size:
        alg_id = u32be_len_input(header["alg"])
        bit_size = key_size
    else:
        alg_id = u32be_len_input(header["enc"])
        bit_size = cek_size

    pub_info = struct.pack(">I", bit_size)
    fixed_info = alg_id + apu_info + apv_info + pub_info

    if tag:
        cctag = u32be_len_input(tag)
        fixed_info += cctag

    ckdf = ConcatKDFHash(
        algorithm=hashes.SHA256(),
        length=bit_size // 8,
        otherinfo=fixed_info,
    )
    return ckdf.derive(shared_key)


def u32be_len_input(s: bytes | str | None, use_base64: bool = False) -> bytes:
    if not s:
        return b"\x00\x00\x00\x00"
    sb: bytes
    if use_base64:
        sb = urlsafe_b64decode(to_bytes(s))
    else:
        sb = to_bytes(s)
    return struct.pack(">I", len(sb)) + sb



import json
from tests.base import TestFixture, load_key
from joserfc.jwk import ECKey, OctKey
from joserfc.jwe import (
    GeneralJSONEncryption,
    FlattenedJSONEncryption,
    encrypt_compact,
    decrypt_compact,
    encrypt_json,
    decrypt_json,
)


payload = (
    b"You can trust us to stick with you through thick and "
    b"thin\xe2\x80\x93to the bitter end. And you can trust us to "
    b"keep any secret of yours\xe2\x80\x93closer than you keep it "
    b"yourself. But you cannot trust us to let you face trouble "
    b"alone, and go off without a word. We are your friends, Frodo."
)


class TestJWERFC7520(TestFixture):
    def run_test(self, data):
        protected = data["protected"]
        key = load_key(data["key"])
        algorithms = [protected["alg"], protected["enc"]]
        value1 = encrypt_compact(protected, payload, key, algorithms=algorithms)
        obj1 = decrypt_compact(value1, key, algorithms=algorithms)
        compact_obj = decrypt_compact(data["compact"], key, algorithms=algorithms)
        self.assertEqual(obj1.protected, compact_obj.protected)
        self.assertEqual(obj1.plaintext, compact_obj.plaintext)

        enc1 = GeneralJSONEncryption(protected, payload)
        enc1.add_recipient(None, key)
        value2 = encrypt_json(enc1, None, algorithms=algorithms)
        obj2 = decrypt_json(value2, key, algorithms=algorithms)
        self.assertEqual(obj2.protected, protected)
        self.assertFalse(obj2.flattened)
        self.assertEqual(obj2.plaintext, payload)

        if "general_json" in data:
            general_obj = decrypt_json(data["general_json"], key, algorithms=algorithms)
            self.assertEqual(general_obj.protected, protected)
            self.assertEqual(general_obj.plaintext, payload)
            self.assertFalse(general_obj.flattened)

        enc2 = FlattenedJSONEncryption(protected, payload)
        enc2.add_recipient(None, key)
        value3 = encrypt_json(enc2, None, algorithms=algorithms)
        obj3 = decrypt_json(value3, key, algorithms=algorithms)
        self.assertEqual(obj3.protected, protected)
        self.assertEqual(obj3.plaintext, payload)
        self.assertTrue(obj3.flattened)
        if "flattened_json" in data:
            flattened_obj = decrypt_json(data["flattened_json"], key, algorithms=algorithms)
            self.assertEqual(flattened_obj.protected, protected)
            self.assertEqual(flattened_obj.plaintext, payload)
            self.assertTrue(flattened_obj.flattened)

    def run_test_agreement(self, data):
        protected = data["protected"]
        ephemeral_key = ECKey.import_key(data["epk"])
        expected_header = {**protected, "epk": ephemeral_key.as_dict(private=False)}
        key = load_key(data["key"])
        algorithms = [protected["alg"], protected["enc"]]
        compact_obj = decrypt_compact(data["compact"], key, algorithms=algorithms)
        self.assertEqual(compact_obj.plaintext, payload)
        self.assertEqual(compact_obj.protected, expected_header)

        general_obj = decrypt_json(data["general_json"], key, algorithms=algorithms)
        self.assertEqual(general_obj.plaintext, payload)
        self.assertEqual(general_obj.protected, expected_header)

        if "flattened_json" in data:
            flattened_obj = decrypt_json(data["flattened_json"], key, algorithms=algorithms)
            self.assertEqual(flattened_obj.plaintext, payload)
            self.assertEqual(flattened_obj.protected, expected_header)

        enc3 = GeneralJSONEncryption(protected, payload)
        enc3.add_recipient(None, key)
        enc3.recipients[0].ephemeral_key = ephemeral_key

        value2 = encrypt_json(enc3, None, algorithms=algorithms)
        obj2 = decrypt_json(value2, key, algorithms=algorithms)
        recipient = obj2.recipients[0]
        self.assertEqual(recipient.headers(), expected_header)

        enc4 = FlattenedJSONEncryption(protected, payload)
        enc4.add_recipient(None, key)
        enc4.recipients[0].ephemeral_key = ephemeral_key
        value3 = encrypt_json(enc4, None, algorithms=algorithms)
        obj3 = decrypt_json(value3, key, algorithms=algorithms)
        recipient = obj3.recipients[0]
        self.assertEqual(recipient.headers(), expected_header)

    def test_5_3(self):
        # Key Wrap Using PBES2-AES-KeyWrap with AES-CBC-HMAC-SHA2
        plaintext = {
            "keys": [
                {
                    "kty": "oct",
                    "kid": "77c7e2b8-6e13-45cf-8672-617b5b45243a",
                    "use": "enc",
                    "alg": "A128GCM",
                    "k": "XctOhJAkA-pD9Lh7ZgW_2A",
                },
                {
                    "kty": "oct",
                    "kid": "81b20965-8332-43d9-a468-82160ad91ac8",
                    "use": "enc",
                    "alg": "A128KW",
                    "k": "GZy6sIZ6wl9NJOKB-jnmVQ",
                },
                {
                    "kty": "oct",
                    "kid": "18ec08e1-bfa9-4d95-b205-2b4dd1d4321d",
                    "use": "enc",
                    "alg": "A256GCMKW",
                    "k": "qC57l_uxcm7Nm3K-ct4GFjx8tM1U8CZ0NLBvdQstiS8",
                },
            ]
        }
        password = OctKey.import_key(b"entrap_o\xe2\x80\x93peter_long\xe2\x80\x93credit_tun")
        protected = {
            "alg": "PBES2-HS512+A256KW",
            "p2s": "8Q1SzinasR3xchYz6ZZcHA",
            "p2c": 8192,
            "cty": "jwk-set+json",
            "enc": "A128CBC-HS256",
        }
        compact_data = """
        eyJhbGciOiJQQkVTMi1IUzUxMitBMjU2S1ciLCJwMnMiOiI4UTFTemluYXNSM3
        hjaFl6NlpaY0hBIiwicDJjIjo4MTkyLCJjdHkiOiJqd2stc2V0K2pzb24iLCJl
        bmMiOiJBMTI4Q0JDLUhTMjU2In0
        .
        d3qNhUWfqheyPp4H8sjOWsDYajoej4c5Je6rlUtFPWdgtURtmeDV1g
        .
        VBiCzVHNoLiR3F4V82uoTQ
        .
        23i-Tb1AV4n0WKVSSgcQrdg6GRqsUKxjruHXYsTHAJLZ2nsnGIX86vMXqIi6IR
        sfywCRFzLxEcZBRnTvG3nhzPk0GDD7FMyXhUHpDjEYCNA_XOmzg8yZR9oyjo6l
        TF6si4q9FZ2EhzgFQCLO_6h5EVg3vR75_hkBsnuoqoM3dwejXBtIodN84PeqMb
        6asmas_dpSsz7H10fC5ni9xIz424givB1YLldF6exVmL93R3fOoOJbmk2GBQZL
        _SEGllv2cQsBgeprARsaQ7Bq99tT80coH8ItBjgV08AtzXFFsx9qKvC982KLKd
        PQMTlVJKkqtV4Ru5LEVpBZXBnZrtViSOgyg6AiuwaS-rCrcD_ePOGSuxvgtrok
        AKYPqmXUeRdjFJwafkYEkiuDCV9vWGAi1DH2xTafhJwcmywIyzi4BqRpmdn_N-
        zl5tuJYyuvKhjKv6ihbsV_k1hJGPGAxJ6wUpmwC4PTQ2izEm0TuSE8oMKdTw8V
        3kobXZ77ulMwDs4p
        .
        0HlwodAhOCILG5SQ2LQ9dg
        """.replace(" ", "").replace("\n", "")
        algorithms = [protected["alg"], protected["enc"]]

        value1 = encrypt_compact(protected, json.dumps(plaintext), password, algorithms=algorithms)
        obj1 = decrypt_compact(value1, password, algorithms=algorithms)
        self.assertEqual(json.loads(obj1.plaintext), plaintext)
        compact_obj = decrypt_compact(compact_data, password, algorithms=algorithms)
        self.assertEqual(json.loads(compact_obj.plaintext), plaintext)


TestJWERFC7520.load_fixture("jwe_rfc7520.json")



from ..registry import Header
from ..errors import MissingCritHeaderError


def is_rfc7797_enabled(header: Header) -> bool:
    if "b64" not in header:
        return False

    if header["b64"] is True:
        return False

    # https://datatracker.ietf.org/doc/html/rfc7797#section-6
    crit = header.get("crit")
    if isinstance(crit, list) and "b64" in crit:
        return True

    raise MissingCritHeaderError("b64")



from __future__ import annotations
import typing as t
from .models import (
    BaseJSONEncryption,
    GeneralJSONEncryption,
    FlattenedJSONEncryption,
    Recipient,
)
from .registry import JWERegistry
from .types import (
    JSONRecipientDict,
    GeneralJSONSerialization,
    FlattenedJSONSerialization,
)
from ..util import (
    to_bytes,
    to_str,
    json_b64encode,
    json_b64decode,
    urlsafe_b64encode,
    urlsafe_b64decode,
)
from .._keys import Key


__all__ = [
    "represent_general_json",
    "represent_flattened_json",
    "extract_general_json",
    "extract_flattened_json",
]


def represent_general_json(obj: GeneralJSONEncryption) -> GeneralJSONSerialization:
    data: GeneralJSONSerialization = __represent_json_serialization(obj)
    recipients = []
    for recipient in obj.recipients:
        item: JSONRecipientDict = {}
        assert recipient.header is not None
        assert recipient.encrypted_key is not None
        item["header"] = recipient.header
        item["encrypted_key"] = to_str(urlsafe_b64encode(recipient.encrypted_key))
        recipients.append(item)
    data["recipients"] = recipients
    return data


def represent_flattened_json(obj: FlattenedJSONEncryption) -> FlattenedJSONSerialization:
    data: FlattenedJSONSerialization = __represent_json_serialization(obj)
    recipient = obj.recipients[0]
    assert recipient is not None
    assert recipient.header is not None
    assert recipient.encrypted_key is not None
    data["header"] = recipient.header
    data["encrypted_key"] = to_str(urlsafe_b64encode(recipient.encrypted_key))
    return data


def __represent_json_serialization(obj: BaseJSONEncryption) -> t.Any:
    data: dict[str, t.Any] = {
        "protected": to_str(json_b64encode(obj.protected)),
        "iv": to_str(obj.base64_segments["iv"]),
        "ciphertext": to_str(obj.base64_segments["ciphertext"]),
        "tag": to_str(obj.base64_segments["tag"]),
    }
    if obj.aad:
        data["aad"] = to_str(urlsafe_b64encode(obj.aad))

    if obj.unprotected:
        data["unprotected"] = obj.unprotected
    return data


def extract_general_json(data: GeneralJSONSerialization, registry: JWERegistry) -> GeneralJSONEncryption:
    protected_segment = to_bytes(data["protected"])
    registry.validate_protected_header_size(protected_segment)
    protected = json_b64decode(protected_segment)

    unprotected = data.get("unprotected")
    base64_segments, bytes_segments, aad = __extract_segments(data, registry)

    obj = GeneralJSONEncryption(protected, None, unprotected, aad)
    obj.base64_segments = base64_segments
    obj.bytes_segments = bytes_segments
    for item in data["recipients"]:
        recipient = __extract_recipient(obj, item, registry)
        obj.recipients.append(recipient)
    return obj


def extract_flattened_json(data: FlattenedJSONSerialization, registry: JWERegistry) -> FlattenedJSONEncryption:
    protected_segment = to_bytes(data["protected"])
    registry.validate_protected_header_size(protected_segment)
    protected = json_b64decode(protected_segment)
    unprotected = data.get("unprotected")
    base64_segments, bytes_segments, aad = __extract_segments(data, registry)
    obj = FlattenedJSONEncryption(protected, None, unprotected, aad)
    obj.base64_segments = base64_segments
    obj.bytes_segments = bytes_segments
    recipient = __extract_recipient(obj, data, registry)
    obj.recipients.append(recipient)
    return obj


def __extract_segments(
    data: t.Union[GeneralJSONSerialization, FlattenedJSONSerialization],
    registry: JWERegistry,
) -> tuple[dict[str, bytes], dict[str, bytes], t.Optional[bytes]]:
    base64_segments: dict[str, bytes] = {
        "iv": to_bytes(data["iv"]),
        "ciphertext": to_bytes(data["ciphertext"]),
        "tag": to_bytes(data["tag"]),
    }
    registry.validate_initialization_vector_size(base64_segments["iv"])
    registry.validate_ciphertext_size(base64_segments["ciphertext"])
    registry.validate_auth_tag_size(base64_segments["tag"])
    bytes_segments: dict[str, bytes] = {
        "iv": urlsafe_b64decode(base64_segments["iv"]),
        "ciphertext": urlsafe_b64decode(base64_segments["ciphertext"]),
        "tag": urlsafe_b64decode(base64_segments["tag"]),
    }
    if "aad" in data:
        aad = urlsafe_b64decode(to_bytes(data["aad"]))
    else:
        aad = None
    return base64_segments, bytes_segments, aad


def __extract_recipient(
    obj: FlattenedJSONEncryption | GeneralJSONEncryption,
    data: FlattenedJSONSerialization | JSONRecipientDict,
    registry: JWERegistry,
) -> Recipient[Key]:
    recipient: Recipient[Key] = Recipient(obj, data.get("header"))
    if "encrypted_key" in data:
        ek_segment = to_bytes(data["encrypted_key"])
        registry.validate_encrypted_key_size(ek_segment)
        recipient.encrypted_key = urlsafe_b64decode(ek_segment)
    return recipient



from unittest import TestCase
from joserfc import jws, jwe, jwt
from joserfc.jwk import OctKey, GuestProtocol, Key, import_key
from joserfc.errors import (
    InvalidPayloadError,
    MissingClaimError,
    UnsupportedHeaderError,
    DecodeError,
)


def use_embedded_jwk(obj: GuestProtocol) -> Key:
    headers = obj.headers()
    return import_key(headers["jwk"])


class TestJWT(TestCase):
    oct_key = OctKey.generate_key()

    def test_default_type(self):
        data = jwt.encode({"alg": "HS256"}, {"sub": "a"}, self.oct_key)
        token = jwt.decode(data, self.oct_key)
        self.assertEqual(token.header["typ"], "JWT")

        data = jwt.encode({"alg": "HS256"}, {"sub": "a"}, self.oct_key, default_type=None)
        token = jwt.decode(data, self.oct_key)
        self.assertNotIn("typ", token.header)

        data = jwt.encode({"alg": "HS256"}, {"sub": "a"}, self.oct_key, default_type="jwt+at")
        token = jwt.decode(data, self.oct_key)
        self.assertEqual(token.header["typ"], "jwt+at")

    def test_invalid_payload(self):
        data = jws.serialize_compact({"alg": "HS256"}, b"hello", self.oct_key)
        self.assertRaises(InvalidPayloadError, jwt.decode, data, self.oct_key)

    def test_claims_registry(self):
        data = jwt.encode({"alg": "HS256"}, {"sub": "a"}, self.oct_key)
        token = jwt.decode(data, self.oct_key)

        claims_registry = jwt.JWTClaimsRegistry(iss={"essential": True})
        self.assertRaises(MissingClaimError, claims_registry.validate, token.claims)

        data = jwt.encode({"alg": "HS256"}, {"iss": "a"}, self.oct_key)
        obj = jwt.decode(data, self.oct_key)
        self.assertEqual(obj.claims["iss"], "a")

    def test_jwe_format(self):
        header = {"alg": "A128KW", "enc": "A128GCM"}
        claims = {"iss": "https://authlib.org"}
        key = OctKey.generate_key(128)
        registry = jwe.JWERegistry()
        result = jwt.encode(header, claims, key, registry=registry)
        self.assertEqual(result.count("."), 4)

        token = jwt.decode(result, key, registry=registry)
        self.assertEqual(token.claims, claims)

    def test_using_registry(self):
        key = OctKey.generate_key(128)
        value1 = jwt.encode({"alg": "HS256"}, {"sub": "a"}, key, registry=jws.JWSRegistry())
        jwt.decode(value1, key, registry=jws.JWSRegistry())
        value2 = jwt.encode({"alg": "A128KW", "enc": "A128GCM"}, {"sub": "a"}, key, registry=jwe.JWERegistry())
        jwt.decode(value2, key, registry=jwe.JWERegistry())

        self.assertRaises(
            KeyError,
            jwt.encode,
            {"alg": "HS256"},
            {"sub": "a"},
            key,
            registry=jwe.JWERegistry(),
        )
        self.assertRaises(
            UnsupportedHeaderError,
            jwt.encode,
            {"alg": "A128KW", "enc": "A128GCM"},
            {"sub": "a"},
            key,
            registry=jws.JWSRegistry(),
        )
        self.assertRaises(
            ValueError,
            jwt.decode,
            value1,
            key,
            registry=jwe.JWERegistry(),
        )
        self.assertRaises(
            DecodeError,
            jwt.decode,
            value2,
            key,
            registry=jws.JWSRegistry(),
        )

    def test_with_embedded_jwk(self):
        value = (
            "eyJqd2siOnsiY3J2IjoiUC0yNTYiLCJ4IjoiVU05ZzVuS25aWFlvdldBbE"
            "03NmNMejl2VG96UmpfX0NIVV9kT2wtZ09vRSIsInkiOiJkczhhZVF3MWwy"
            "Y0RDQTdiQ2tPTnZ3REtwWEFidFhqdnFDbGVZSDhXc19VIiwia3R5IjoiRU"
            "MifSwiYWxnIjoiRVMyNTYifQ.eyJpc3MiOiJ1cm46ZXhhbXBsZTppc3N1Z"
            "XIiLCJhdWQiOiJ1cm46ZXhhbXBsZTphdWRpZW5jZSIsImlhdCI6MTYwNDU"
            "4MDc5NH0.60boak3_dErnW47ZPty1C0nrjeVq86EN_eK0GOq6K8w2OA0th"
            "KoBxFK4j-NuU9yZ_A9UKGxPT_G87DladBaV9g"
        )
        token = jwt.decode(value, use_embedded_jwk)
        self.assertEqual(token.claims["iss"], "urn:example:issuer")



from unittest import TestCase
from joserfc import jws
from joserfc.jwk import RSAKey, OctKey
from joserfc.registry import HeaderParameter
from joserfc.errors import (
    BadSignatureError,
    MissingKeyError,
    UnsupportedAlgorithmError,
    UnsupportedKeyUseError,
    UnsupportedKeyAlgorithmError,
    UnsupportedKeyOperationError,
    InvalidKeyTypeError,
    InvalidKeyCurveError,
    MissingHeaderError,
    MissingCritHeaderError,
    UnsupportedHeaderError,
    InvalidHeaderValueError,
)
from joserfc.util import urlsafe_b64encode
from tests.base import load_key


class TestJWSErrors(TestCase):
    key = OctKey.import_key("secret")

    def test_without_alg(self):
        self.assertRaises(MissingHeaderError, jws.serialize_compact, {"kid": "123"}, "i", self.key)

    def test_raise_unsupported_algorithm_error(self):
        registry = jws.JWSRegistry(algorithms=["HS256", "HS384", "HS512"])
        header = {"alg": "HS256"}
        jws.serialize_compact(header, "i", self.key, registry=registry)
        # raise error
        registry = jws.JWSRegistry(algorithms=["HS512"])
        self.assertRaises(UnsupportedAlgorithmError, jws.serialize_compact, header, "i", self.key, registry=registry)

    def test_without_key(self):
        self.assertRaises(MissingKeyError, jws.serialize_compact, {"alg": "HS256"}, "i", None)

        header = {"alg": "HS256"}
        text = jws.serialize_compact(header, "i", self.key)
        self.assertRaises(MissingKeyError, jws.deserialize_compact, text, None)

    def test_none_alg(self):
        header = {"alg": "none"}
        text = jws.serialize_compact(header, "i", None, algorithms=["none"])
        obj = jws.deserialize_compact(text, None, algorithms=["none"])
        self.assertEqual(obj.payload, b"i")
        # none alg has no signature
        text += "aQ"
        self.assertRaises(BadSignatureError, jws.deserialize_compact, text, None, algorithms=["none"])

    def test_header_invalid_type(self):
        # kid should be a string
        header = {"alg": "HS256", "kid": 123}
        self.assertRaises(
            InvalidHeaderValueError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )

        # jwk should be a dict
        header = {"alg": "HS256", "jwk": "dict"}
        self.assertRaises(
            InvalidHeaderValueError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )

        # jku should be a URL
        header = {"alg": "HS256", "jku": "url"}
        self.assertRaises(
            InvalidHeaderValueError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )

        # x5c should be a chain of string
        header = {"alg": "HS256", "x5c": "url"}
        self.assertRaises(
            InvalidHeaderValueError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )
        header = {"alg": "HS256", "x5c": [1, 2]}
        self.assertRaises(
            InvalidHeaderValueError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )

    def test_crit_header(self):
        header = {"alg": "HS256", "crit": ["kid"]}
        # missing kid header
        self.assertRaises(
            MissingCritHeaderError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )

        header = {"alg": "HS256", "kid": "1", "crit": ["kid"]}
        jws.serialize_compact(header, "i", self.key)

    def test_unsupported_crit_header(self):
        header = {"alg": "HS256", "bob": "a", "crit": ["bob"]}
        self.assertRaises(
            UnsupportedHeaderError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )

        registry = jws.JWSRegistry(
            header_registry={
                "bob": HeaderParameter("Bob", "str"),
            }
        )
        # allow with custom header registry
        jws.serialize_compact(header, "i", self.key, registry=registry)

    def test_extra_header(self):
        header = {"alg": "HS256", "extra": "hi"}
        self.assertRaises(
            UnsupportedHeaderError,
            jws.serialize_compact,
            header,
            "i",
            self.key,
        )
        # bypass extra header
        registry = jws.JWSRegistry(strict_check_header=False)
        jws.serialize_compact(header, "i", self.key, registry=registry)

        # or use a header registry
        extra_header = {"extra": HeaderParameter("Extra header", "str", False)}
        registry = jws.JWSRegistry(header_registry=extra_header)
        jws.serialize_compact(header, "i", self.key, registry=registry)

    def test_rsa_invalid_signature(self):
        key1 = RSAKey.generate_key()
        key2 = RSAKey.generate_key()
        header = {"alg": "RS256"}
        text = jws.serialize_compact(header, "i", key1)
        self.assertRaises(BadSignatureError, jws.deserialize_compact, text, key2)

        header = {"alg": "PS256"}
        text = jws.serialize_compact(header, "i", key1, algorithms=["PS256"])
        self.assertRaises(BadSignatureError, jws.deserialize_compact, text, key2, algorithms=["PS256"])

    def test_ec_incorrect_curve(self):
        header = {"alg": "ES256"}
        key = load_key("ec-p512-private.pem")
        self.assertRaises(InvalidKeyCurveError, jws.serialize_compact, header, "i", key)

    def test_ec_invalid_signature(self):
        header = {"alg": "ES256"}
        key1 = load_key("ec-p256-alice.json")
        key2 = load_key("ec-p256-bob.json")
        text = jws.serialize_compact(header, "i", key1)
        self.assertRaises(BadSignatureError, jws.deserialize_compact, text, key2)

        parts = text.split(".")
        bad_text = ".".join(parts[:-1]) + "." + urlsafe_b64encode(b"abc").decode("utf-8")
        self.assertRaises(BadSignatureError, jws.deserialize_compact, bad_text, key1)

    def test_okp_bad_signature(self):
        header = {"alg": "EdDSA"}
        key1 = load_key("okp-ed448-private.pem")
        key2 = load_key("okp-ed25519-private.json")
        algorithms = ["EdDSA"]
        value = jws.serialize_json({"protected": header}, "i", key1, algorithms=algorithms)
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_json,
            value,
            key2,
            algorithms=algorithms,
        )


class TestJWSWithKeyErrors(TestCase):
    def test_invalid_key_use(self):
        key = OctKey.generate_key(parameters={"use": "enc"})
        header = {"alg": "HS256"}
        self.assertRaises(UnsupportedKeyUseError, jws.serialize_compact, header, "i", key)

    def test_invalid_key_alg(self):
        key = OctKey.generate_key(parameters={"alg": "HS512"})
        header = {"alg": "HS256"}
        self.assertRaises(UnsupportedKeyAlgorithmError, jws.serialize_compact, header, "i", key)

    def test_invalid_key_ops(self):
        key = OctKey.generate_key(parameters={"key_ops": ["verify"]})
        header = {"alg": "HS256"}
        self.assertRaises(UnsupportedKeyOperationError, jws.serialize_compact, header, "i", key)

    def test_invalid_key_type(self):
        key = OctKey.generate_key()
        header = {"alg": "RS256"}
        self.assertRaises(InvalidKeyTypeError, jws.serialize_compact, header, "i", key)
